from typing import Callable

import numpy as np
from ..base import Transformer


class Observable(Transformer):
    r""" An object that transforms a series of state vectors :math:`X\in\mathbb{R}^{T\times n}` into a
    series of observables :math:`\Psi(X)\in\mathbb{R}^{T\times k}`, where `n` is the dimension of each state vector
    and `k` the dimension of the observable space.
    """

    def _evaluate(self, x: np.ndarray):
        r""" Evalues the observable on input data `x`.

        Parameters
        ----------
        x : (T, n) ndarray
            Input data.

        Returns
        -------
        y : (T, m) ndarray
            Basis applied to input data.
        """
        raise NotImplementedError()

    def __call__(self, x: np.ndarray):
        r""" Evaluation of the observable.

        Parameters
        ----------
        x : (N, d) np.ndarray
            Evaluates the observable for N d-dimensional data points.

        Returns
        -------
        out : (N, p) np.ndarray
            Result of the evaluation for each data point.
        """
        return self._evaluate(x)

    def transform(self, data, **kwargs):
        return self(data)


class Concatenation(Observable):
    r"""Concatenation operation to evaluate :math:`(f_1 \circ f_2)(x) = f_1(f_2(x))`, where
    :math:`f_1` and :math:`f_2` are observables.

    Parameters
    ----------
    obs1 : Callable
        First observable :math:`f_1`.
    obs2 : Callable
        Second observable :math:`f_2`.
    """

    def __init__(self, obs1: Callable[[np.ndarray], np.ndarray], obs2: Callable[[np.ndarray], np.ndarray]):
        self.obs1 = obs1
        self.obs2 = obs2

    def _evaluate(self, x: np.ndarray) -> np.ndarray:
        return self.obs1(self.obs2(x))



import numbers
from typing import Optional, Union, Callable

import numpy as np
from threadpoolctl import threadpool_limits

from ..base import Estimator
from ..numeric import is_sorted, spd_inv_sqrt, schatten_norm
from ..util.parallel import joining


def vamp_score(koopman_model, r: Union[float, str],
               covariances_test=None, dim: Optional[int] = None, epsilon: float = 1e-10):
    """Compute the VAMP score between a covariance-based Koopman model and potentially a
    test model for cross-validation.

    Parameters
    ----------
    koopman_model : deeptime.decomposition.CovarianceKoopmanModel
        The model to score.
    r : float or str
        The type of score to evaluate. Can by an floating point value greater or equal to 1 or 'E', yielding the
        VAMP-r score or the VAMP-E score, respectively. :footcite:`wu2020variational`
        Typical choices (also accepted as inputs) are:

        *  'VAMP1'  Sum of singular values of the half-weighted Koopman matrix.
                    If the model is reversible, this is equal to the sum of
                    Koopman matrix eigenvalues, also called Rayleigh quotient :footcite:`wu2020variational`.
        *  'VAMP2'  Sum of squared singular values of the half-weighted Koopman
                    matrix :footcite:`wu2020variational`. If the model is reversible, this is
                    equal to the kinetic variance :footcite:`noe2015kinetic`.
        *  'VAMPE'  Approximation error of the estimated Koopman operator with respect to
                    the true Koopman operator up to an additive constant :footcite:`wu2020variational` .

    covariances_test : deeptime.covariance.CovarianceModel, optional, default=None

        If `test_model` is not None, this method computes the cross-validation score
        between self and `covariances_test`. It is assumed that self was estimated from
        the "training" data and `test_model` was estimated from the "test" data. The
        score is computed for one realization of self and `test_model`. Estimation
        of the average cross-validation score and partitioning of data into test and
        training part is not performed by this method.

        If `covariances_test` is None, this method computes the VAMP score for the model
        contained in self.

    dim : int, optional, default=None
        Artificially restrict the scoring to the top `dim` slowest processes.

    epsilon : float, default=1e-10


    Returns
    -------
    score : float
        If `test_model` is not None, returns the cross-validation VAMP score between
        self and `test_model`. Otherwise return the selected VAMP-score of self.

    Notes
    -----
    If the Koopman model was estimated using correlations that are based on data with its sample mean removed,
    this effectively removes the constant function from the singular function space and artificially lowers the score
    by 1. This is accounted for in this method, i.e., if :code:`koopman_model.cov.data_mean_removed` evaluates to
    `True`, the score is internally incremented by 1.

    The VAMP-:math:`r` and VAMP-E scores are computed according to :footcite:`wu2020variational`,
    Equation (33) and Equation (30), respectively.

    References
    ----------
    .. footbibliography::
    """
    if dim is not None:
        dim = min(koopman_model.koopman_matrix.shape[0], dim)
    if isinstance(r, str):
        r = r.lower()
        r = r.replace("vamp", "")
        if r.isnumeric():
            r = float(r)
        else:
            assert r == 'e', "only VAMP-E supported, otherwise give as float >= 1"
    else:
        assert isinstance(r, numbers.Number) and r >= 1, "score only for r >= 1 or r = \"E\""
    if covariances_test is None:
        cov_test = koopman_model.cov
    else:
        cov_test = covariances_test
    assert koopman_model.cov.data_mean_removed == cov_test.data_mean_removed, \
        "Covariances must be consistent with respect to the data"
    if koopman_model.cov.cov_00.shape != cov_test.cov_00.shape:
        raise ValueError(f"Shape mismatch, the covariances had "
                         f"shapes {koopman_model.cov.cov_00.shape} and {cov_test.cov_00.shape}.")
    if not is_sorted(koopman_model.singular_values, 'desc'):
        sort_ix = np.argsort(koopman_model.singular_values)[::-1][:dim]  # indices to sort in descending order
    else:
        sort_ix = np.arange(koopman_model.singular_values.shape[0])[:dim]  # already sorted

    U = koopman_model.instantaneous_coefficients[:, sort_ix]
    V = koopman_model.timelagged_coefficients[:, sort_ix]

    if r == 'e':
        K = np.diag(koopman_model.singular_values[sort_ix])
        # see https://arxiv.org/pdf/1707.04659.pdf eqn. (30)
        score = np.trace(2.0 * np.linalg.multi_dot([K, U.T, cov_test.cov_0t, V])
                         - np.linalg.multi_dot([K, U.T, cov_test.cov_00, U, K, V.T, cov_test.cov_tt, V]))
    else:
        # see https://arxiv.org/pdf/1707.04659.pdf eqn. (33)
        A = np.atleast_2d(spd_inv_sqrt(U.T.dot(cov_test.cov_00).dot(U), epsilon=epsilon))
        B = np.atleast_2d(U.T.dot(cov_test.cov_0t).dot(V))
        C = np.atleast_2d(spd_inv_sqrt(V.T.dot(cov_test.cov_tt).dot(V), epsilon=epsilon))
        ABC = np.linalg.multi_dot([A, B, C])
        score = schatten_norm(ABC, r) ** r
    if koopman_model.cov.data_mean_removed:
        score += 1
    return score


def vamp_score_data(data, data_lagged, transformation=None, r=2, epsilon=1e-6, dim=None):
    r""" Computes VAMP score based on data and corresponding time-lagged data.
    Can be equipped with a transformation, defaults to 'identity' transformation.

    Parameters
    ----------
    data : (T, n) ndarray
        Instantaneous data.
    data_lagged : (T, n) ndarray
        Time-lagged data.
    transformation : Callable
        Transformation on data that will be scored.
    r : int or str, optional, default=2
        The type of VAMP score evaluated, see :meth:`deeptime.decomposition.vamp_score`.
    epsilon : float, optional, default=1e-6
        Regularization parameter for the score, see :meth:`deeptime.decomposition.vamp_score`.
    dim : int, optional, default=None
        Number of components that should be scored. Defaults to all components. See
        :meth:`deeptime.decomposition.vamp_score`.

    Returns
    -------
    score : float
        The VAMP score.

    See Also
    --------
    vamp_score
    """
    if transformation is None:
        def transformation(x):
            return x
    from deeptime.decomposition import VAMP
    model = VAMP(epsilon=epsilon, observable_transform=transformation).fit((data, data_lagged)).fetch_model()
    return model.score(r=r, dim=dim, epsilon=epsilon)


def blocksplit_trajs(trajs, lag=1, sliding=True, shift=None, random_state=None):
    """ Splits trajectories into approximately uncorrelated fragments.

    Will split trajectories into fragments of lengths lag or longer. These fragments
    are overlapping in order to conserve the transition counts at given lag.
    If sliding=True, the resulting trajectories will lead to exactly the same count
    matrix as when counted from dtrajs. If sliding=False (sampling at lag), the
    count matrices are only equal when also setting shift=0.

    Parameters
    ----------
    trajs : list of ndarray(int)
        Trajectories
    lag : int
        Lag time at which counting will be done.
    sliding : bool
        True for splitting trajectories for sliding count, False if lag-sampling will be applied
    shift : None or int
        Start of first full tau-window. If None, shift will be randomly generated
    random_state : None or int or np.random.RandomState
        Random seed to use.

    Returns
    -------
    blocks : list of ndarray
        The blocks.
    """
    from sklearn.utils.random import check_random_state
    random_state = check_random_state(random_state)
    blocks = []
    for traj in trajs:
        if len(traj) <= lag:
            continue
        if shift is None:
            s = random_state.randint(min(lag, traj.size - lag))
        else:
            s = shift
        if sliding:
            if s > 0:
                blocks.append(traj[:lag + s])
            for t0 in range(s, len(traj) - lag, lag):
                blocks.append(traj[t0:t0 + 2 * lag])
        else:
            for t0 in range(s, len(traj) - lag, lag):
                blocks.append(traj[t0:t0 + lag + 1])
    return blocks


def cvsplit_trajs(trajs, random_state=None):
    """ Splits the trajectories into a training and test set with approximately equal number of trajectories

    Parameters
    ----------
    trajs : list of ndarray(int)
        Discrete trajectories
    random_state : None or int or np.random.RandomState
        Random seed to use.
    """
    from sklearn.utils.random import check_random_state
    assert len(trajs) > 1, 'Only have a single trajectory. Cannot be split into train and test set'
    random_state = check_random_state(random_state)
    I0 = random_state.choice(len(trajs), int(len(trajs) / 2), replace=False)
    I1 = np.array(list(set(list(np.arange(len(trajs)))) - set(list(I0))))
    train_set = [trajs[i] for i in I0]
    test_set = [trajs[i] for i in I1]
    return train_set, test_set


def vamp_score_cv(fit_fetch: Union[Estimator, Callable], trajs, lagtime=None, n=10, splitting_mode="sliding", r=2,
                  dim: Optional[int] = None, blocksplit: bool = True, random_state=None, n_jobs=1):
    r""" Scores the MSM using the variational approach for Markov processes and cross-validation.

    Implementation and ideas following :footcite:`noe2013variational` :footcite:`wu2020variational` and
    cross-validation :footcite:`mcgibbon2015variational`.

    Divides the data into training and test data, fits a MSM using the training
    data using the parameters of this estimator, and scores is using the test
    data.
    Currently only one way of splitting is implemented, where for each n,
    the data is randomly divided into two approximately equally large sets of
    discrete trajectory fragments with lengths of at least the lagtime.

    Currently only implemented using dense matrices - will be slow for large state spaces.

    Parameters
    ----------
    fit_fetch : callable or estimator
        Can be provided as callable for a custom fit and fetch method. Should be a function pointer or lambda which
        takes a list of discrete trajectories as input and yields a
        :class:`CovarianceKoomanModel <deeptime.decomposition.CovarianceKoopmanModel>`. Or an estimator which
        yields this kind of model.
    trajs : list of array_like
        Input data.
    lagtime : int, optional, default=None
        lag time, must be provided if blocksplitting is used, otherwise can be left None
    splitting_mode : str, optional, default="sliding"
        Can be one of "sliding" and "sample". In former case the blocks may overlap, otherwise not.
    n : number of samples
        Number of repetitions of the cross-validation. Use large n to get solid means of the score.
    r : float or str, default=2
        Available scores are based on the variational approach for Markov processes :footcite:`noe2013variational`
        :footcite:`wu2020variational`, see :meth:`deeptime.decomposition.vamp_score` for available options.
    blocksplit : bool, optional, default=True
        Whether to perform blocksplitting (see :meth:`blocksplit_dtrajs` ) before evaluating folds. Defaults to `True`.
        In case no blocksplitting is performed, individual dtrajs are used for training and validation. This means that
        at least two dtrajs must be provided (`len(dtrajs) >= 2`), otherwise this method raises an exception.
    dim : int or None, optional, default=None
        The maximum number of eigenvalues or singular values used in the score. If set to None,
        all available eigenvalues will be used.
    random_state : None or int or np.random.RandomState
        Random seed to use.
    n_jobs : int, optional, default=1
        Number of jobs for folds. In case n_jobs is 1, no parallelization.

    References
    ----------
    .. footbibliography::
    """
    from deeptime.util.parallel import handle_n_jobs
    from deeptime.util.types import ensure_timeseries_data

    if blocksplit and lagtime is None:
        raise ValueError("In case blocksplit is used, please provide a lagtime.")

    n_jobs = handle_n_jobs(n_jobs)
    if isinstance(fit_fetch, Estimator):
        fit_fetch = _FitFetch(fit_fetch)

    ttrajs = ensure_timeseries_data(trajs)  # ensure format
    if splitting_mode not in ('sliding', 'sample'):
        raise ValueError('vamp_score_cv currently only supports count modes "sliding" and "sample"')
    scores = np.empty((n,), float)
    sliding = splitting_mode == 'sliding'

    if random_state is None or isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)
    assert isinstance(random_state, np.random.RandomState)

    args = [(i, fit_fetch, ttrajs, r, dim, lagtime, blocksplit, sliding, random_state, n_jobs) for i in range(n)]

    if n_jobs > 1:
        from multiprocessing import get_context
        with joining(get_context("spawn").Pool(processes=n_jobs)) as pool:
            for result in pool.imap_unordered(_worker, args):
                fold, score = result
                scores[fold] = score
    else:
        for fold in range(n):
            _, score = _worker(args[fold])
            scores[fold] = score
    return scores


class _FitFetch:

    def __init__(self, est):
        self._est = est

    def __call__(self, x):
        return self._est.fit(x).fetch_model()


def _worker(args):
    from deeptime.markov.msm import MarkovStateModel
    fold, fit_fetch, ttrajs, r, dim, lagtime, blocksplit, sliding, random_state, n_jobs = args

    with threadpool_limits(limits=1 if n_jobs > 1 else None, user_api='blas'):
        if blocksplit:
            trajs_split = blocksplit_trajs(ttrajs, lag=lagtime, sliding=sliding,
                                           random_state=random_state)
        else:
            trajs_split = ttrajs
            assert len(trajs_split) > 1, "Need at least two trajectories if blocksplit " \
                                         "is not used to decompose the data."
        trajs_train, trajs_test = cvsplit_trajs(trajs_split, random_state=random_state)
        # this is supposed to construct a markov state model from data directly, for example what fit_fetch could do is
        train_model = fit_fetch(trajs_train)
        if isinstance(train_model, MarkovStateModel):
            score = train_model.score(trajs_test, r=r, dim=dim)
        else:
            test_model = fit_fetch(trajs_test)
            score = train_model.score(r=r, test_model=test_model, dim=dim)
        return fold, score



r"""
.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
"""

import numpy as np
import scipy.sparse

from . import objective_sparse

from .linsolve import mydot
from .linsolve_sparse import factor_aug as factor
from .linsolve_sparse import solve_factorized_aug as solve_factorized

__all__ = ['solve_mle_rev', ]

"""Algorithm parameters"""
GAMMA_MIN = 0.0001
GAMMA_MAX = 0.01
GAMMA_BAR = 0.49
KAPPA = 0.01
TAU = 0.5
RHO = min(0.2, min((0.5 * GAMMA_BAR) ** (1.0 / TAU), 1.0 - KAPPA))
SIGMA = 0.1
BETA = 100000


def mynorm(x):
    return np.linalg.norm(x)


def wrap_function(function, args):
    ncalls = [0]
    if function is None:
        return ncalls, None

    def function_wrapper(*wrapper_args):
        ncalls[0] += 1
        return function(*(wrapper_args + args))

    return ncalls, function_wrapper


def primal_dual_solve(func, x0, Dfunc, A, b, G, h, args=(), tol=1e-10,
                      maxiter=100, full_output=False):
    """Wrap calls to function and Jacobian"""
    fcalls, func = wrap_function(func, args)
    Dfcalls, Dfunc = wrap_function(Dfunc, args)

    M, N = G.shape
    P, N = A.shape

    """Total number of inequality constraints"""
    m = M

    def gap(z):
        r"""Gap-function"""
        l = z[N + P:N + P + M]
        s = z[N + P + M:]
        return mydot(l, s) / m

    def centrality(z):
        r"""Centrality function"""
        l = z[N + P:N + P + M]
        s = z[N + P + M:]
        return np.min(l * s)

    def KKT(z, sigma=0.0):
        r"""KKT system (possible perturbed)."""

        """Primal variable"""
        x = z[0:N]

        """Multiplier for equality constraints"""
        nu = z[N:N + P]

        """Multiplier for inequality constraints"""
        l = z[N + P:N + P + M]

        """Slacks"""
        s = z[N + P + M:]

        """Evaluate objective function"""
        F = func(x)

        """Dual infeasibility"""
        rdual = F + mydot(A.transpose(), nu) + mydot(G.transpose(), l)

        """Primal infeasibilities"""
        rprim1 = mydot(A, x) - b
        rprim2 = mydot(G, x) - h + s

        """Complementary slackness (perturbed)"""
        mu = gap(z)
        rslack = l * s - sigma * mu

        return np.hstack((rdual, rprim1, rprim2, rslack))

    def step_fast(z, KKTval, LU, G, A, mu, beta, gamma, alpha0):
        r"""Affine scaling step."""
        dz = solve_factorized(z, KKTval, LU, G, A)

        """Reduce step length until slacks s and multipliers l are positive"""
        alpha = 1.0 * alpha_0
        k = 0
        for k in range(10):
            z_new = z + alpha * dz
            if np.all(z_new[N + P:] > 0.0):
                break
            alpha *= 0.5
            k += 1
        if k == 10 - 1:
            raise RuntimeError("Maximum steplength reduction reached")

        """Reduce step length until iterates lie in correct neighborhood"""
        for k in range(10):
            z_new = z + alpha * dz
            KKTval_new = KKT(z_new)
            dual = mynorm(KKTval_new[0:N])
            prim = mynorm(KKTval_new[N:N + P + M])
            mu_new = gap(z_new)
            cent_new = centrality(z_new)

            if (dual <= beta * mu_new and prim <= beta * mu_new and
                    cent_new >= gamma * mu_new):
                break
            alpha *= 0.5
            # alpha *= 0.95
        if k == 10 - 1:
            raise RuntimeError("Maximum steplength reduction reached")
        return z_new, mu_new

    def step_safe(z, KKTval, LU, G, A, mu, beta, gamma):
        r"""Centering step."""
        dz = solve_factorized(z, KKTval, LU, G, A)

        """Reduce step length until slacks s and multipliers l are positive"""
        alpha = 1.0
        k = 0
        for k in range(10):
            z_new = z + alpha * dz
            if np.all(z_new[N + P:] > 0.0):
                break
            alpha *= 0.5
            k += 1
        if k == 10 - 1:
            raise RuntimeError("Maximum steplength reduction (pos.) reached")

        """Reduce step length until iterates lie in correct neighborhood
        and mu fulfills Armijo condition"""
        k = 0
        for k in range(10):
            z_new = z + alpha * dz
            KKTval_new = KKT(z_new, sigma=SIGMA)
            dual = mynorm(KKTval_new[0:N])
            prim = mynorm(KKTval_new[N:N + P + M])
            mu_new = gap(z_new)
            cent_new = centrality(z_new)

            if (dual <= beta * mu_new and prim <= beta * mu_new and
                    cent_new >= gamma * mu_new and
                    mu_new <= (1.0 - KAPPA * alpha * (1.0 - SIGMA)) * mu):
                break
            alpha *= 0.5
        if k == 10 - 1:
            raise RuntimeError("Maximum steplength reduction reached")
        return z_new, mu_new

    """INITIALIZATION"""

    """Initial Slacks for inequality constraints"""
    s0 = -1.0 * (mydot(G, x0) - h)

    """Initial multipliers for inequality constraints"""
    l0 = 1.0 * np.ones(M)

    """Initial multipliers for equality constraints"""
    nu0 = np.zeros(P)

    """Initial point"""
    z0 = np.hstack((x0, nu0, l0, s0))

    """Initial KKT-values"""
    KKTval0 = KKT(z0, sigma=0.0)
    mu0 = gap(z0)
    dual0 = mynorm(KKTval0[0:N])
    prim0 = mynorm(KKTval0[N:N + P + M])

    """Initial neighborhood"""
    beta = BETA * np.sqrt(dual0 ** 2 + prim0 ** 2) / mu0
    gamma = 1.0 * GAMMA_MAX

    """Number of fast steps"""
    t = 0

    """Number of iterations"""
    n = 0

    """Dummy variable for step type"""
    step_type = " "

    """MAIN LOOP"""
    z = z0
    x = z0[0:N]
    KKTval = KKTval0
    dual = dual0
    prim = prim0
    mu = mu0
    Dfunc_val = Dfunc(x)
    LU = factor(z, Dfunc_val, G, A)

    if full_output:
        info = {'z': []}
        info['z'].append(z)

    for n in range(maxiter):
        """Attempt fast step"""
        beta_new = (1.0 + GAMMA_BAR ** (t + 1)) * beta
        gamma_new = GAMMA_MIN + GAMMA_BAR ** (t + 1) * (GAMMA_MAX - GAMMA_MIN)
        alpha_0 = 1.0 - np.sqrt(mu) / GAMMA_BAR ** t

        if alpha_0 > 0.0:
            z_new, mu_new = step_fast(z, KKTval, LU, G, A, mu,
                                      beta_new, gamma_new, alpha_0)
            if mu_new < RHO * mu:
                """Fast successful"""
                z = z_new
                mu = mu_new
                beta = beta_new
                gamma = gamma_new
                t += 1
                step_type = "f"
            else:
                """Perturbed right-had side"""
                KKTval_pert = 1.0 * KKTval
                KKTval_pert[N + P + M:] -= SIGMA * mu
                z, mu = step_safe(z, KKTval_pert, LU, G, A, mu, beta, gamma)
                step_type = "s"
        else:
            """Perturbed right-hand side"""
            KKTval_pert = 1.0 * KKTval
            KKTval_pert[N + P + M:] -= SIGMA * mu
            z, mu = step_safe(z, KKTval_pert, LU, G, A, mu, beta, gamma)
            step_type = "s"

        """Compute new iterates"""
        KKTval = KKT(z, sigma=0.0)
        dual = mynorm(KKTval[0:N])
        prim = mynorm(KKTval[N:N + P + M])
        x = z[0:N]
        Dfunc_val = Dfunc(x)
        LU = factor(z, Dfunc_val, G, A)
        if full_output:
            info['z'].append(z)
        if mu < tol and dual < tol and prim < tol:
            break
    if n == maxiter - 1:
        raise RuntimeError("Maximum number of iterations reached")

    if full_output:
        return z[0:N], info
    else:
        return z[0:N]


def solve_mle_rev(C, tol=1e-10, maxiter=100, full_output=False,
                  return_statdist=True):
    """Number of states"""
    M = C.shape[0]

    """Initial guess for primal-point"""
    z0 = np.zeros(2 * M)
    z0[0:M] = 1.0

    """Inequality constraints"""
    # G = np.zeros((M, 2*M))
    # G[np.arange(M), np.arange(M)] = -1.0
    G = -1.0 * scipy.sparse.eye(M, n=2 * M, k=0)
    h = np.zeros(M)

    """Equality constraints"""
    A = np.zeros((1, 2 * M))
    A[0, M] = 1.0
    b = np.array([0.0])

    """Scaling"""
    c0 = C.max()
    C = C / c0

    """Symmetric part"""
    Cs = C + C.T

    """Column sum"""
    c = C.sum(axis=0)

    if scipy.sparse.issparse(C):
        Cs = Cs.tocsr()
        c = c.A1
        A = scipy.sparse.csr_matrix(A)
        F = objective_sparse.F
        DF = objective_sparse.DFsym
        convert_solution = objective_sparse.convert_solution
    else:
        raise ValueError("dense not supported")

    """PDIP iteration"""
    res = primal_dual_solve(F, z0, DF, A, b, G, h,
                            args=(Cs, c),
                            maxiter=maxiter, tol=tol,
                            full_output=full_output)
    if full_output:
        z, info = res
    else:
        z = res
    pi, P = convert_solution(z, Cs)

    result = [P]
    if return_statdist:
        result.append(pi)
    if full_output:
        result.append(info)

    return tuple(result) if len(result) > 1 else result[0]



from ._vampnet import VAMPNet, VAMPNetModel, koopman_matrix, sym_inverse, covariances, vamp_score, vampnet_loss
from ._tae import TAE, TAEModel, TVAE, TVAEModel, TVAEEncoder
from ._kvadnet import kvad_score



from typing import Optional, Callable

import numpy as np
import scipy

from ._koopman import TransferOperatorModel
from ..base import Model, Transformer, EstimatorTransformer
from ..kernels import Kernel
from ..numeric import sort_eigs, eigs
from ..util.types import to_dataset


class DMDModel(Model, Transformer):
    r""" Model produced by the :class:`DMD` estimator.

    Parameters
    ----------
    eigenvalues : (n,) ndarray
        The DMD eigenvalues.
    modes : (n, n) ndarray
        The DMD modes.
    mode : str, optional, default='exact'
        The mode of estimation that was used. See :attr:`DMD.available_modes`.
    """

    def __init__(self, eigenvalues: np.ndarray, modes: np.ndarray, mode='exact'):
        super().__init__()
        self.eigenvalues = eigenvalues
        self.modes = modes
        self.mode = mode

    def transform(self, data: np.ndarray, **kwargs):
        r""" Transforms an input trajectory by applying the model's captured dynamics.

        Parameters
        ----------
        data : (T, n) ndarray
            Input trajectory
        **kwargs
            Compatibility.

        Returns
        -------
        output_trajectory : (T, n) ndarray

        """
        if self.mode == 'exact':
            modes_adj = np.linalg.pinv(self.modes.T)
            return np.linalg.multi_dot([
                self.modes.T, np.diag(self.eigenvalues), modes_adj, data.T
            ]).T
        else:
            return np.linalg.multi_dot([
                self.modes.T, np.diag(self.eigenvalues), self.modes.conj(), data.T
            ]).T


class DMD(EstimatorTransformer):
    r""" Dynamic mode decomposition estimator. :footcite:`schmid2010dynamic`

    There are two supported modes:

    * `standard`, which produces "projected" DMD modes (following the original formulation of DMD),
    * `exact`, which produces DMD modes that do not required ordered data but just
      matched pairs of data :footcite:`tu2013dynamic`.

    Parameters
    ----------
    mode : str
        The estimation mode, see :attr:`available_modes` for available modes.
    rank : int or None, optional, default=None
        Truncation of the rank after performing SVD.
    driver : str, default='numpy'
        Which package to use for the SVD. Defaults to numpy, can also be 'scipy'.

    Notes
    -----
    In standard DMD, one considers a temporally ordered list of
    data vectors :math:`(z_0,\ldots,z_T)\in\mathbb{R}^{T\times d}`. The data is split into the pair

    .. math::

        X = (z_0, \ldots, z_{T-1}),\quad Y=(z_1,\ldots, z_T).

    If the mode is `exact`, the list does not need to be temporally ordered but just the pairs :math:`(X_i, Y_i)`
    have to match. The underlying assumption is that the data are generated by a linear relationship

    .. math::

        z_{t+1} = A z_t

    for some matrix :math:`A`.

    The so-called DMD modes and eigenvalues are then the (potentially scaled) eigenvectors and eigenvalues of :math:`A`.

    References
    ----------
    .. footbibliography::
    """

    available_modes = 'exact', 'standard'  #: The available estimation modes.
    available_drivers = 'numpy', 'scipy'  #: The available drivers.

    def __init__(self, mode='exact', rank=None, driver='scipy'):
        super().__init__()
        if driver not in DMD.available_drivers:
            raise ValueError(f"Invalid driver {driver}, must be one of {DMD.available_drivers}.")
        self.mode = mode
        self.rank = rank
        self.driver = driver

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        if value not in DMD.available_modes:
            raise ValueError(f"Invalid mode {value}, must be one of {DMD.available_modes}.")
        self._mode = value

    def _svd(self, mat, **kw):
        if self.driver == 'numpy':
            return np.linalg.svd(mat, **kw)
        elif self.driver == 'scipy':
            return scipy.linalg.svd(mat, **kw)

    def _eig(self, mat, **kw):
        if self.driver == 'numpy':
            return np.linalg.eig(mat, **kw)
        elif self.driver == 'scipy':
            return scipy.linalg.eig(mat, **kw)

    def fit(self, data, **kwargs):
        r""" Fit this estimator instance onto data.

        Parameters
        ----------
        data
            Input data, see :meth:`to_dataset <deeptime.util.types.to_dataset>` for options.
        **kwargs
            Kwargs, may contain lagtime.

        Returns
        -------
        self : DMD
            Reference to self.
        """
        dataset = to_dataset(data, lagtime=kwargs.get("lagtime", None))
        X, Y = dataset[:]
        X, Y = X.T, Y.T  # per convention arrays are [T, d] so here we transpose them

        U, s, Vt = self._svd(X, full_matrices=False)
        if self.rank is not None:
            rank = min(self.rank, U.shape[1])
            U = U[:, :rank]
            s = s[:rank]
            Vt = Vt[:rank]
        V = Vt.conj().T
        S_inv = np.diag(1 / s)
        A = np.linalg.multi_dot([U.conj().T, Y, V, S_inv])

        eigenvalues, eigenvectors = self._eig(A)
        eigenvalues, eigenvectors = sort_eigs(eigenvalues, eigenvectors, order='lexicographic')

        if self.mode == 'exact':
            dmd_modes = np.linalg.multi_dot([Y, V, S_inv, eigenvectors, np.diag(1 / eigenvalues)])
        elif self.mode == 'standard':
            dmd_modes = U @ eigenvectors

        self._model = DMDModel(eigenvalues, dmd_modes.T)
        return self

    def fetch_model(self) -> Optional[DMDModel]:
        r""" Yields the estimated model if :meth:`fit` was called.

        Returns
        -------
        model : DMDModel or None
            The model or None.
        """
        return self._model

    def transform(self, data, **kwargs):
        r""" See :meth:`DMDModel.transform`.

        Parameters
        ----------
        data : (T, d) np.ndarray
            Input data

        Returns
        -------
        result : (T, d) np.ndarray
            Propagated input data
        """
        return self.fetch_model().transform(data, **kwargs)


class EDMDModel(TransferOperatorModel):
    r""" The EDMD model which can be estimated from a :class:`EDMD` estimator. It possesses the estimated operator
    as well as capabilities to project onto modes.

    Parameters
    ----------
    operator : (n, n) ndarray
        The estimated operator.
    basis : callable
        The basis transform that was used.
    eigenvalues : (n, ) ndarray
        Eigenvalues for the modes.
    modes : (k, n) ndarray
        The EDMD modes.

    See Also
    --------
    EDMD
    """

    def __init__(self, operator: np.ndarray, basis: Callable[[np.ndarray], np.ndarray], eigenvalues, modes):
        super().__init__(operator, instantaneous_obs=basis, timelagged_obs=basis)
        self.basis = basis
        self.eigenvalues = eigenvalues
        self.modes = modes
        self.n_eigenvalues = len(self.eigenvalues)

    def transform(self, data: np.ndarray, **kw):
        r"""Transforms the data by first applying the basis and then the estimated modes, i.e.,

        .. math::

            X \mapsto \Psi(X)\varphi

        where :math:`X` is the input data, :math:`\Psi` the basis transform, and :math:`\varphi` the Koopman operator's

        Parameters
        ----------
        data : (T, n) ndarray
            Input data

        Returns
        -------
        transformed : (T, n) ndarray
            The forward transform of the input data.
        """
        modes = self.modes[:self.n_eigenvalues]
        out = self.instantaneous_obs(data)
        return out @ modes.T


class EDMD(EstimatorTransformer):
    r""" Extended dynamic mode decomposition for estimation of the Koopman (or optionally Perron-Frobenius)
    operator. :footcite:`williams2015data` :footcite:`klus2016edmd`.

    The estimator needs a basis :math:`\Psi : \mathbb{R}^n\to\mathbb{R}^k, \mathbf{x}\mapsto\Psi(\mathbf{x}))`
    and data matrices :math:`X = [x_1,\ldots,x_M]`, :math:`Y=[y_1,\ldots,y_M]` of time-lagged pairs of data.
    It then estimates a Koopman operator approximation :math:`K` so that :math:`\Psi(y_i)\approx K^\top \Psi(x_i)`.

    In other words, for data matrices :math:`\Psi_X` and :math:`\Psi_Y` it solves the minimization problem

    .. math::

        \min\| \Psi_Y - K\Psi_X\|_F.

    Parameters
    ----------
    basis : callable
        The basis callable, maps from (T, k) ndarray to (T, m) ndarray. See :mod:`deeptime.basis` for a selection
        of pre-defined bases.
    n_eigs : int, optional, default=None
        The number of eigenvalues, determining the number of dominant singular functions / modes being estimated.
        If None, estimates all eigenvalues / eigenvectors.
    operator : str, default='koopman'
        Which operator to estimate, see :attr:`available_operators`.

    References
    ----------
    .. footbibliography::
    """

    available_operators = 'koopman', 'perron-frobenius'  #: The supported operators.

    def __init__(self, basis: Callable[[np.ndarray], np.ndarray], n_eigs: Optional[int] = None,
                 operator: str = 'koopman'):
        super().__init__()
        self.basis = basis
        self.operator = operator
        self.n_eigs = n_eigs

    def fetch_model(self) -> Optional[EDMDModel]:
        r""" Yields the estimated model or None.

        Returns
        -------
        model : EDMDKoopmanModel or None
            The model.
        """
        return self._model

    def fit(self, data, **kwargs):
        r""" Fit this estimator instance onto data.

        Parameters
        ----------
        data
            Input data, see :meth:`to_dataset <deeptime.util.types.to_dataset>` for options.
        **kwargs
            Kwargs, may contain lagtime.

        Returns
        -------
        self : EDMD
            Reference to self.
        """
        dataset = to_dataset(data, lagtime=kwargs.get("lagtime", None))
        x, y = dataset[:]
        n_data = x.shape[0]
        assert n_data == y.shape[0], "Trajectories for data and timelagged data must be of same length!"
        psi_x = self.basis(x).T
        psi_y = self.basis(y).T

        cov_00 = (1/x.shape[0]) * psi_x @ psi_x.T
        cov_0t = (1/x.shape[0]) * psi_x @ psi_y.T

        if self.operator != 'koopman':
            cov_0t = cov_0t.T

        m_edmd = scipy.linalg.pinv(cov_00) @ cov_0t
        eig_val, eig_vec = eigs(m_edmd, self.n_eigs)
        eig_val, eig_vec = sort_eigs(eig_val, eig_vec, order='lexicographic')
        self._model = EDMDModel(m_edmd, basis=self.basis, eigenvalues=eig_val, modes=eig_vec.T)
        return self


class KernelEDMDModel(TransferOperatorModel):
    r""" The kEDMD model containing eigenvalues and eigenfunctions evaluated in the instantaneous data.

    Parameters
    ----------
    eigenvalues : (d,) ndarray
        The eigenvalues.
    eigenvectors : (T, d) ndarray
        The eigenfunction evaluation.
    kernel : Kernel
        The kernel that was used for estimation.

    See Also
    --------
    KernelEDMD
    """

    def __init__(self, data: np.ndarray, eigenvalues: np.ndarray, eigenvectors: np.ndarray, kernel: Kernel):
        super().__init__(eigenvectors @ np.diag(eigenvalues),
                         instantaneous_obs=lambda x: self.kernel.apply(x, self.data),
                         timelagged_obs=lambda x: self.kernel.apply(x, self.data))
        self.data = data
        self.eigenvalues = eigenvalues
        self.eigenvectors = eigenvectors
        self.kernel = kernel

    def transform(self, data: np.ndarray, propagate=True):
        return self.instantaneous_obs(data) @ self.eigenvectors


class KernelEDMD(EstimatorTransformer):
    r""" Estimator implementing kernel extended mode decomposition. :footcite:`williams2016kernel`
    :footcite:`klus2019eigendecomposition` :footcite:`klus2018kernel`.

    Parameters
    ----------
    kernel : Kernel
        The kernel to use. See :mod:`deeptime.kernels` for a list of available kernels.
    epsilon : float, optional, default=0
        Regularization parameter.
    n_eigs : int, optional, default=None
        Number of eigenvalue/eigenvector pairs to compute.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, kernel: Kernel, epsilon: float = 0., n_eigs: Optional[int] = None):
        super().__init__()
        self.kernel = kernel
        self.epsilon = epsilon
        self.n_eigs = n_eigs

    def fetch_model(self) -> Optional[KernelEDMDModel]:
        r""" Yields the estimated model or `None`.

        Returns
        -------
        model : KernelEDMDModel or None
            The model.
        """
        return super().fetch_model()

    def fit(self, data, **kwargs):
        r""" Fit this estimator instance onto data.

        Parameters
        ----------
        data
            Input data, see :meth:`to_dataset <deeptime.util.types.to_dataset>` for options.
        **kwargs
            Kwargs, may contain lagtime.

        Returns
        -------
        self : KernelEDMD
            Reference to self.
        """
        dataset = to_dataset(data, lagtime=kwargs.get("lagtime", None))
        x, y = dataset[:]
        gram_0 = self.kernel.gram(x)  # G_XX
        gram_1 = self.kernel.apply(x, y)  # G_XY

        A = scipy.linalg.solve(gram_0 + self.epsilon * np.eye(gram_0.shape[0]), gram_1.T, assume_a='sym')
        eigenvalues, eigenvectors = eigs(A, n_eigs=self.n_eigs)
        eigenvalues, eigenvectors = sort_eigs(eigenvalues, eigenvectors)
        self._model = KernelEDMDModel(x, eigenvalues, eigenvectors, self.kernel)
        return self



from typing import Optional
from warnings import warn

import numpy as np
from scipy.integrate import odeint
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import LinearRegression, ridge_regression
from sklearn.metrics import r2_score
from sklearn.preprocessing import PolynomialFeatures
from sklearn.utils.validation import check_X_y

from ..base import Estimator, Model
from ..numeric import drop_nan_rows


class SINDyModel(Model):
    r"""The SINDy model. Stores the parameters learned by a :class:`SINDy` estimator
    to encode a first order differential equation model for the measurement data.
    It can be used to make derivative predictions, simulate forward in time from
    initial conditions, and for self-scoring.

    The model encodes a dynamical system

    .. math::
            \dot{X} = \Theta(X)\Xi

    via the following correpondences
    :code:`library` = :math:`\Theta` and
    :code:`(intercept, coefficients)` = :math:`\Xi^\top`.

    Parameters
    ----------
    library : library object
        The feature library, :math:`\Theta`.
        It is assumed that this object has already been fit to the input data.
        The object should implement  :meth:`transform`
        and :meth:`get_feature_names` methods.

    coefficients : np.ndarray, shape (n_input_features, n_output_features)
        Coefficients giving the linear combination of basis functions to
        approximate derivative data, i.e. :math:`\Xi^\top`. Note that
        :code:`coefficients` may or may not contain information about the intercepts
        depending on the library used (e.g. a polynomial library can contain the
        constant function).

    input_features : iterable of str
        List of input feature names.

    intercept : float, optional, default=0
        The intercept/bias for the learned model.
        If the library already contains a constant function, there is no
        need for an intercept term.
    """

    def __init__(self, library, coefficients, input_features, intercept=0):
        super().__init__()
        self.library = library
        self.coef_ = coefficients
        self.input_features = input_features
        self.intercept_ = intercept

    def transform(self, x):
        r"""Apply the functions of the feature library.

        This method computes :math:`\Theta(X)`.

        Parameters
        ----------
        x : np.ndarray, shape (n_samples, n_input_features)
            Measurement data.

        Returns
        -------
        y : np.ndarray, shape (n_samples, n_output_features)
            The feature library evaluated on :code:`x`,
            i.e. :math:`\Theta(X)`.
        """
        return self.library.transform(x)

    def predict(self, x):
        r"""Predict the derivative of :code:`x` using the learned model,
        i.e. predict :math:`\dot{X}` via :math:`\dot{X} \approx \Theta(X)\Xi`.

        Parameters
        ----------
        x : np.ndarray, shape (n_samples, n_input_features)
            Measurement data.

        Returns
        -------
        y : np.ndarray, shape (n_samples, n_input_features)
            Model prediction of the derivative :math:`\dot{X}`.
        """
        return np.matmul(self.transform(x), self.coef_.T) + self.intercept_

    def score(self, x, y=None, t=None, scoring=r2_score, **scoring_kws):
        r"""Compute a score for the time derivative prediction produced by the model.

        Parameters
        ----------
        x : np.ndarray, shape (n_samples, n_input_features)
            Measurement data.

        y : np.ndarray, shape (n_samples, n_input_features), optional, default=None
            Array of derivatives of :code:`x`.
            By default, :code:`np.gradient` is used to compute the derivatives.

        t : np.ndarray, shape (n_samples,) or a scalar, optional, default=None
            The times when the measurements in :code:`x` were taken or the
            (uniform) time spacing between measurements in :code:`x`.
            By default a timestep of 1 is assumed.
            This argument is ignored if :code:`y` is passed.

        scoring : callable, optional, default=r2_score
            Function by which to score the prediction.
            By default, the R^22 coefficient of determination is computed.
            See `Scikit-learn <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
            for more options.

        Returns
        -------
        s : float
            Score for the time derivative prediction.
        """
        if y is None:
            if t is None:
                y = np.gradient(x, axis=0)
            else:
                y = np.gradient(x, t, axis=0)

        return scoring(y, self.predict(x), **scoring_kws)

    def simulate(self, x0, t, integrator=odeint, integrator_kws=None):
        """
        Simulate the SINDy model forward in time.

        Parameters
        ----------
        x0: numpy array, size [n_features]
            Initial condition from which to simulate.

        t: int or numpy array of size [n_samples]
            If the model is in continuous time, t must be an array of time
            points at which to simulate. If the model is in discrete time,
            t must be an integer indicating how many steps to predict.

        integrator: callable, optional (default :code:`odeint`)
            Function to use to integrate the system.
            Default is :code:`scipy.integrate.odeint`.

        integrator_kws: dict, optional (default None)
            Optional keyword arguments to pass to the integrator

        Returns
        -------
        x: numpy array, shape (n_samples, n_features)
            Simulation results.
        """

        if integrator_kws is None:
            integrator_kws = {}

        def rhs(x, _):
            return self.predict(x[np.newaxis, :])[0]

        return integrator(rhs, x0, t, **integrator_kws)

    def print(self, lhs=None, precision=3):
        r"""Print the learned equations in a human-readable way.

        Parameters
        ----------
        lhs : list of strings, optional, default=None
            List of variables to print on the left-hand side of the equations.
            By default :code:`self.input_features` are used.

        precision: int, optional, default=3
            Precision to be used when printing out model coefficients.
        """
        equations = self.equations(precision=precision)
        for i, eqn in enumerate(equations):
            if lhs:
                print(f"{lhs[i]}' = {eqn}")
            else:
                print(f"{self.input_features[i]}' = {eqn}")

    def equations(self, precision=3):
        """
        Get the right-hand sides of the learned equations.

        Parameters
        ----------
        precision : int, optional, default=3
            Precision to which coefficients are rounded.

        Returns
        -------
        equations : list of strings, length (n_input_features)
            List of model equations, with one for each input variable.
        """
        feature_names = self.library.get_feature_names(
            input_features=self.input_features
        )
        equation_list = [None] * len(self.coef_)

        for k, row_coef in enumerate(self.coef_):
            terms = [
                "{coef:.{precision}f} {fn}".format(coef=coef, precision=precision, fn=feature_names[i])
                for i, coef in enumerate(row_coef)
                if coef != 0
            ]
            equation_list[k] = " + ".join(terms)

        return equation_list

    @property
    def coefficients(self):
        r"""Returns the learned model coefficients, i.e. :math:`\Xi^\top`

        Returns
        -------
        coefficients : np.ndarray, shape (n_input_features, n_output_features)
            Learned model coefficients :math:`\Xi^\top`.
        """
        return self.coef_

    @property
    def intercept(self):
        r"""Returns the intercept (bias) for the learned model.

        Returns
        -------
        intercept : np.ndarray, shape (n_input_features,) or float
            The intercept or intercepts for each input feature.
        """
        return self.intercept_


class SINDy(Estimator):
    r"""Learn a dynamical systems model for measurement data using the
    Sparse Identification of Nonlinear Dynamical Systems (SINDy) method.

    For given measurement data :math:`X`, and a set of library functions evaluated
    on :math:`X`

    .. math::
        \Theta(X) = [\theta_1(X), \theta_2(X), \dots, \theta_k(X)],

    SINDy seeks a sparse set of coefficients :math:`\Xi` which satisfies

    .. math::
        \dot{X} \approx \Theta(X)\Xi.

    The i-th column of this matrix equation gives a differential equation for the
    i-th measurement variable (i-th column in :math:`X`). For more details see
    :footcite:`brunton2016sindy`.

    Parameters
    ----------
    library : library object, optional, default=None
        The candidate feature library, :math:`\Theta`.
        The object should implement a :meth:`fit`, :meth:`transform`,
        and :meth:`get_feature_names` methods. It should also have
        :attr:`n_input_features_` and :attr:`n_output_features_` attributes.
        By default a polynomial library of degree 2 is used.

    optimizer : optimizer object, optional, default=None
        The optimization routine used to solve the objective
        :math:`\dot{X} \approx \Theta(X)\Xi`.
        The object should have :meth:`fit` and :meth:`predict` methods
        and :attr:`coef_` and :attr:`intercept_` attributes. For example,
        any linear regressor from `sklearn.linear_model \
        <https://scikit-learn.org/stable/modules/linear_model.html>`_ should work.
        By default, :meth:`STLSQ` is used.

    input_features : list of strings, optional, default=None
        List of input feature names. By default, the names
        "x0", "x1", ..., "x{n_input_features}" is used.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, library=None, optimizer=None, input_features=None):
        super().__init__()
        if library is None:
            library = PolynomialFeatures(degree=2)
        if optimizer is None:
            optimizer = STLSQ(threshold=0.1)
        self.library = library
        self.optimizer = optimizer
        self.input_features = input_features

    def fit(self, x, y=None, t=None):
        r"""Fit the estimator to measurement data.

        Parameters
        ----------
        x : np.ndarray, shape (n_samples, n_input_features)
            Training/measurement data.
            Each row should correspond to one example and each column
            to a feature.

        y : np.ndarray, shape (n_samples, n_input_features), optional, default=None
            Array of derivatives of :code:`x`.
            By default, :code:`np.gradient` is used to compute the derivatives.

        t : np.ndarray, shape (n_samples,) or a scalar, optional, default=None
            The times when the measurements in :code:`x` were taken or the
            (uniform) time spacing between measurements in :code:`x`.
            By default a timestep of 1 is assumed.
            This argument is ignored if :code:`y` is passed.

        Returns
        -------
        self: SINDy
            Reference to self

        """
        if y is not None:
            x_dot = y
        elif t is not None:
            x_dot = np.gradient(x, t, axis=0)
        else:
            x_dot = np.gradient(x, axis=0)

        # Some differentiation methods produce nans near boundaries
        x, x_dot = drop_nan_rows(x, x_dot)

        steps = [("features", self.library), ("model", self.optimizer)]
        if hasattr(self.library, 'fit'):
            self.library.fit(x)

        self.n_input_features_ = x.shape[1]
        x_transformed = self.library.transform(x)
        self.n_output_features_ = x_transformed[1]
        self.optimizer.fit(x_transformed, x_dot)
        if self.input_features is None:
            self.input_features = [f"x{i}" for i in range(self.n_input_features_)]

        if hasattr(self.optimizer, "intercept_"):
            intercept = self.optimizer.intercept_
        else:
            intercept = 0

        self._model = SINDyModel(
            library=self.library,
            coefficients=self.optimizer.coef_,
            input_features=self.input_features,
            intercept=intercept,
        )

        return self

    def fetch_model(self) -> Optional[SINDyModel]:
        r""" Yields the latest model.

        Returns
        -------
        model : SINDyModel or None
            The model.
        """
        return super().fetch_model()


class STLSQ(LinearRegression):
    r"""Sequentially thresholded least squares algorithm.

    Attempts to minimize the objective function
    :math:`\|y - Xw\|^2_2 + \alpha \|w\|^2_2`
    by iteratively performing least squares and masking out
    elements of the weight that are below a given threshold.

    See this paper for more details :footcite:`brunton2016sindy`.

    Parameters
    ----------
    threshold : float, optional, default=0.1
        Minimum magnitude for a coefficient in the weight vector.
        Coefficients with magnitude below the threshold are set
        to zero.

    alpha : float, optional, default=0.05
        Optional L2 (ridge) regularization on the weight vector.

    max_iter : int, optional, default=20
        Maximum iterations of the optimization algorithm.

    ridge_kw : dict, optional, default=None
        Optional keyword arguments to pass to the ridge regression.

    fit_intercept : boolean, optional, default=False
        Whether to calculate the intercept for this model. If set to false, no
        intercept will be used in calculations.

    normalize : boolean, optional, default=False
        This parameter is ignored when :code:`fit_intercept` is set to False. If True,
        the regressors X will be normalized before regression by subtracting
        the mean and dividing by the l2-norm.

    copy_X : boolean, optional, default=True
        If True, X will be copied; else, it may be overwritten.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, threshold=0.1, alpha=0.05, max_iter=20, ridge_kw=None, normalize=False, fit_intercept=False,
                 copy_X=True):
        super().__init__(fit_intercept=fit_intercept, normalize=normalize, copy_X=copy_X)
        self.threshold = threshold
        self.alpha = alpha
        self.max_iter = max_iter
        self.ridge_kw = ridge_kw
        self.normalize = normalize
        self.fit_intercept = fit_intercept
        self.copy_X = copy_X

    def fit(self, x_, y):
        r"""Fit to the data.

        Parameters
        ----------
        x_ : array-like, shape (n_samples, n_features)
            Training data (:math:`X` in the above equation).

        y : array-like, shape (n_samples,) or (n_samples, n_targets)
            Target values (:math:`y` in the above equation).

        Returns
        -------
        self : STLSQ
            Reference to self
        """
        # Do some preprocessing before fitting
        x_, y = check_X_y(x_, y, accept_sparse=[], y_numeric=True, multi_output=True)

        x, y, X_offset, y_offset, X_scale = self._preprocess_data(
            x_,
            y,
            fit_intercept=self.fit_intercept,
            normalize=self.normalize,
            copy=self.copy_X,
        )

        self.iters = 0
        self.ind_ = np.ones((y.shape[1], x.shape[1]), dtype=bool)
        self.coef_ = np.linalg.lstsq(x, y, rcond=None)[0].T  # initial guess
        self.history_ = [self.coef_]

        self._reduce(x, y)
        self.ind_ = np.abs(self.coef_) > 1e-14

        self._set_intercept(X_offset, y_offset, X_scale)
        return self

    def _reduce(self, x, y):
        r"""Iterates the thresholding. Assumes an initial guess is saved in
        self.coef_ and self.ind_
        """
        ind = self.ind_
        n_samples, n_features = x.shape
        n_targets = y.shape[1]
        n_features_selected = np.sum(ind)

        for _ in range(self.max_iter):
            if np.count_nonzero(ind) == 0:
                warn(
                    "Sparsity parameter is too big ({}) and eliminated all "
                    "coefficients".format(self.threshold)
                )
                coef = np.zeros((n_targets, n_features))
                break

            coef = np.zeros((n_targets, n_features))
            for i in range(n_targets):
                if np.count_nonzero(ind[i]) == 0:
                    warn(
                        "Sparsity parameter is too big ({}) and eliminated all "
                        "coefficients".format(self.threshold)
                    )
                    continue
                coef_i = self._regress(x[:, ind[i]], y[:, i])
                coef_i, ind_i = self._sparse_coefficients(
                    n_features, ind[i], coef_i, self.threshold
                )
                coef[i] = coef_i
                ind[i] = ind_i

            self.history_.append(coef)
            if np.sum(ind) == n_features_selected or self._no_change():
                # could not (further) select important features
                break
        else:
            warn(
                "STLSQ._reduce did not converge after {} iterations.".format(
                    self.max_iter
                ),
                ConvergenceWarning,
            )
            try:
                coef
            except NameError:
                coef = self.coef_
                warn(
                    "STLSQ._reduce has no iterations left to determine coef",
                    ConvergenceWarning,
                )
        self.coef_ = coef
        self.ind_ = ind

    @staticmethod
    def _sparse_coefficients(dim, ind, coef, threshold):
        r"""Perform thresholding of the weight vector(s)"""
        c = np.zeros(dim)
        c[ind] = coef
        big_ind = np.abs(c) >= threshold
        c[~big_ind] = 0
        return c, big_ind

    def _regress(self, x, y):
        r"""Perform the ridge regression"""
        kw = self.ridge_kw or {}
        coef = ridge_regression(x, y, self.alpha, **kw)
        self.iters += 1
        return coef

    def _no_change(self):
        r"""Check if the coefficient mask has changed after thresholding"""
        this_coef = self.history_[-1].flatten()
        if len(self.history_) > 1:
            last_coef = self.history_[-2].flatten()
        else:
            last_coef = np.zeros_like(this_coef)
        return all(bool(i) == bool(j) for i, j in zip(this_coef, last_coef))



import numpy as _np
from scipy.sparse import csr_matrix
from scipy.sparse.base import issparse

from deeptime.util.sparse import remove_negative_entries

__docformat__ = "restructuredtext en"
__author__ = "Benjamin Trendelkamp-Schroer, Martin Scherer, Moritz Hoffmann, Frank Noe"
__copyright__ = "Copyright 2014, Computational Molecular Biology Group, FU-Berlin"
__credits__ = ["Benjamin Trendelkamp-Schroer", "Martin Scherer", "Moritz Hoffmann", "Frank Noe"]


# ======================================================================
# Flux matrix operations
# ======================================================================

def flux_matrix(T, pi, qminus, qplus, netflux=True):
    r"""Compute the TPT flux network for the reaction A-->B.

    Parameters
    ----------
    T : (M, M) ndarray
        transition matrix
    pi : (M,) ndarray
        Stationary distribution corresponding to T
    qminus : (M,) ndarray
        Backward comittor
    qplus : (M,) ndarray
        Forward committor
    netflux : boolean
        True: net flux matrix will be computed
        False: gross flux matrix will be computed

    Returns
    -------
    flux : (M, M) ndarray
        Matrix of flux values between pairs of states.

    See also
    --------
    committor.forward_committor, committor.backward_committor

    Notes
    -----
    Computation of the flux network relies on transition path theory
    (TPT) [1]_. Here we use discrete transition path theory [2]_ in
    the transition matrix formulation [3]_.
    The central object used in transition path theory is the
    forward and backward comittor function.

    The TPT (gross) flux is defined as

    .. math::
        f_{ij}=\left \{ \begin{array}{rl}
            \pi_i q_i^{(-)} p_{ij} q_j^{(+)} & i \neq j \\
            0                                & i=j
        \end{array} \right .

    The TPT net flux is then defined as

    .. math:: f_{ij}=\max\{f_{ij} - f_{ji}, 0\} \:\:\:\forall i,j.

    References
    ----------
    .. [1] W. E and E. Vanden-Eijnden.
        Towards a theory of transition paths.
        J. Stat. Phys. 123: 503-523 (2006)
    .. [2] P. Metzner, C. Schuette and E. Vanden-Eijnden.
        Transition Path Theory for Markov Jump Processes.
        Multiscale Model Simul 7: 1192-1219 (2009)
    .. [3] F. Noe, Ch. Schuette, E. Vanden-Eijnden, L. Reich and
        T. Weikl: Constructing the Full Ensemble of Folding Pathways
        from Short Off-Equilibrium Simulations.
        Proc. Natl. Acad. Sci. USA, 106, 19011-19016 (2009)
    """
    if issparse(T):
        flux = T.multiply(qplus[None, ...]).multiply(pi[..., None]).multiply(qminus[..., None])
        # Remove self-fluxes
        flux.setdiag(0)
        flux.eliminate_zeros()
    else:
        flux = pi[:, None] * qminus[:, None] * T * qplus[None, :]
        # Remove self fluxes
        flux[_np.diag_indices(T.shape[0])] = 0.0

    # Return net or gross flux
    if netflux:
        return to_netflux(flux)
    else:
        return flux


def to_netflux(flux):
    r"""Compute the netflux from the gross flux.

    Parameters
    ----------
    flux : (M, M) ndarray
        Matrix of flux values between pairs of states.

    Returns
    -------
    netflux : (M, M) ndarray
        Matrix of netflux values between pairs of states.

    Notes
    -----
    The netflux or effective current is defined as

    .. math::

        f_{ij}^{+}=\max \{ f_{ij}-f_{ji}, 0 \},

    see [1]_.

    :math:`f_{ij}` is the flux for the transition from :math:`A` to
    :math:`B`.

    References
    ----------
    .. [1] P. Metzner, C. Schuette and E. Vanden-Eijnden.
        Transition Path Theory for Markov Jump Processes.
        Multiscale Model Simul 7: 1192-1219 (2009)

    """
    if issparse(flux):
        netflux = remove_negative_entries(flux - flux.T)
    else:
        netflux = _np.maximum(0, flux - flux.T)
    return netflux


def flux_production(F):
    r"""Returns the net flux production for all states

    Parameters
    ----------
    F : (M, M) ndarray
        Matrix of flux values between pairs of states.

    Returns
    -------
    prod : (M,) ndarray
        Array containing flux production (positive) or consumption
        (negative) at each state

    """
    influxes = _np.array(F.sum(axis=0)).flatten()  # all that flows in
    outfluxes = _np.array(F.sum(axis=1)).flatten()  # all that flows out
    prod = outfluxes - influxes  # net flux into nodes
    return prod


def _flux_producers_consumers(F, rtol, atol, producers):
    influxes = _np.array(_np.sum(F, axis=0)).flatten()  # all that flows in
    outfluxes = _np.array(_np.sum(F, axis=1)).flatten()  # all that flows out
    # net out flux absolute
    if producers:
        net_abs = _np.maximum(outfluxes - influxes, 0)
    else:
        net_abs = _np.maximum(influxes - outfluxes, 0)
    # net out flux relative
    prod_rel = net_abs / (_np.maximum(outfluxes, influxes))
    # return all indexes that are produces in terms of absolute and relative tolerance
    return list(_np.where((net_abs > atol) * (prod_rel > rtol))[0])


def flux_producers(F, rtol=1e-05, atol=1e-12):
    r"""Return indexes of states that are net flux producers.

    Parameters
    ----------
    F : (M, M) ndarray
        Matrix of flux values between pairs of states.
    rtol : float
        relative tolerance. fulfilled if max(outflux-influx, 0) / max(outflux,influx) < rtol
    atol : float
        absolute tolerance. fulfilled if max(outflux-influx, 0) < atol

    Returns
    -------
    producers : (M, ) ndarray of int
        indices of states that are net flux producers. May include
        "dirty" producers, i.e.  states that have influx but still
        produce more outflux and thereby violate flux conservation.

    """
    return _flux_producers_consumers(F, rtol, atol, producers=True)


def flux_consumers(F, rtol=1e-05, atol=1e-12):
    r"""Return indexes of states that are net flux producers.

    Parameters
    ----------
    F : (M, M) ndarray
        Matrix of flux values between pairs of states.
    rtol : float
        relative tolerance. fulfilled if max(outflux-influx, 0) / max(outflux,influx) < rtol
    atol : float
        absolute tolerance. fulfilled if max(outflux-influx, 0) < atol

    Returns
    -------
    producers : (M, ) ndarray of int
        indexes of states that are net flux producers. May include
        "dirty" producers, i.e.  states that have influx but still
        produce more outflux and thereby violate flux conservation.

    """
    return _flux_producers_consumers(F, rtol, atol, producers=False)


def coarsegrain(F, sets):
    r"""Coarse-grains the flux to the given sets. See [1]_ .

    Parameters
    ----------
    F : (n, n) ndarray or scipy.sparse matrix
        Matrix of flux values between pairs of states.
    sets : list of array-like of ints
        The sets of states onto which the flux is coarse-grained.

    Notes
    -----
    The coarse grained flux is defined as

    .. math:: fc_{I,J} = \sum_{i \in I,j \in J} f_{i,j}

    Note that if you coarse-grain a net flux, it does n ot necessarily
    have a net flux property anymore. If want to make sure you get a
    netflux, use to_netflux(coarsegrain(F,sets)).

    References
    ----------
    .. [1] F. Noe, Ch. Schuette, E. Vanden-Eijnden, L. Reich and
        T. Weikl: Constructing the Full Ensemble of Folding Pathways
        from Short Off-Equilibrium Simulations.
        Proc. Natl. Acad. Sci. USA, 106, 19011-19016 (2009)

    """
    nnew = len(sets)
    if issparse(F):
        Fc = csr_matrix((nnew, nnew))
        Fin = F.tocsr()
    else:
        Fc = _np.zeros((nnew, nnew))
        Fin = F

    for i in range(0, nnew - 1):
        for j in range(i + 1, nnew):
            I = list(sets[i])
            J = list(sets[j])
            Fc[i, j] = (Fin[I, :][:, J]).sum()
            Fc[j, i] = (Fin[J, :][:, I]).sum()
    return Fc


# ======================================================================
# Total flux, rate and mfpt for the A->B reaction
# ======================================================================


def total_flux(F, A=None):
    r"""Compute the total flux, or turnover flux, that is produced by
        the flux sources and consumed by the flux sinks. [1]_

    Parameters
    ----------
    F : (M, M) ndarray
        Matrix of flux values between pairs of states.
    A : array_like (optional)
        List of integer state labels for set A (reactant)

    Returns
    -------
    F : float
        The total flux, or turnover flux, that is produced by the flux
        sources and consumed by the flux sinks

    References
    ----------
    .. [1] P. Metzner, C. Schuette and E. Vanden-Eijnden.
        Transition Path Theory for Markov Jump Processes.
        Multiscale Model Simul 7: 1192-1219 (2009)

    """
    if A is None:
        prod = flux_production(F)
        outflux = _np.sum(_np.maximum(prod, 0))
        return outflux
    else:
        X = set(_np.arange(F.shape[0]))  # total state space
        A = set(A)
        notA = X.difference(A)

        if issparse(F):
            # Extract rows corresponding to A
            W = F.tocsr()
            W = W[list(A), :]
            # Extract columns corresponding to X\A
            W = W.tocsc()
            W = W[:, list(notA)]

            F = W.sum()
            return F
        else:
            outflux = (F[list(A), :])[:, list(notA)].sum()
            return outflux


def rate(totflux, pi, qminus):
    r"""Transition rate for reaction A to B.

    Parameters
    ----------
    totflux : float
        The total flux between reactant and product
    pi : (M,) ndarray
        Stationary distribution
    qminus : (M,) ndarray
        Backward comittor

    Returns
    -------
    kAB : float
        The reaction rate (per time step of the
        Markov chain)

    See also
    --------
    committor, total_flux, flux_matrix

    Notes
    -----
    Computation of the rate relies on discrete transition path theory
    (TPT). The transition rate, i.e. the total number of reaction events per
    time step, is given in [1]_ as:

    .. math:: k_{AB}=\frac{1}{F} \sum_i \pi_i q_i^{(-)}

    :math:`F` is the total flux for the transition from :math:`A` to
    :math:`B`.

    References
    ----------
    .. [1] F. Noe, Ch. Schuette, E. Vanden-Eijnden, L. Reich and
        T. Weikl: Constructing the Full Ensemble of Folding Pathways
        from Short Off-Equilibrium Simulations.
        Proc. Natl. Acad. Sci. USA, 106, 19011-19016 (2009)

    """
    kAB = totflux / (pi * qminus).sum()
    return kAB


def mfpt(totflux, pi, qminus):
    r"""Mean first passage time for reaction A to B.

    Parameters
    ----------
    totflux : float
        The total flux between reactant and product
    pi : (M,) ndarray
        Stationary distribution
    qminus : (M,) ndarray
        Backward comittor

    Returns
    -------
    tAB : float
        The mean first-passage time for the A to B reaction

    See also
    --------
    rate

    Notes
    -----
    Equal to the inverse rate, see [1]_.

    References
    ----------
    .. [1] F. Noe, Ch. Schuette, E. Vanden-Eijnden, L. Reich and
        T. Weikl: Constructing the Full Ensemble of Folding Pathways
        from Short Off-Equilibrium Simulations.
        Proc. Natl. Acad. Sci. USA, 106, 19011-19016 (2009)

    """
    return 1.0 / rate(totflux, pi, qminus)


###############################################################################
# Pathway decomposition
###############################################################################

def pathways(F, A, B, fraction=1.0, maxiter=1000):
    r"""Decompose flux network into dominant reaction paths. [1]_

    Parameters
    ----------
    F : (M, M) scipy.sparse matrix
        The flux network (matrix of netflux values)
    A : array_like
        The set of starting states
    B : array_like
        The set of end states
    fraction : float, optional
        Fraction of total flux to assemble in pathway decomposition
    maxiter : int, optional
        Maximum number of pathways for decomposition

    Returns
    -------
    paths : list
        List of dominant reaction pathways
    capacities: list
        List of capacities corresponding to each reactions pathway in paths

    Notes
    -----
    The default value for fraction is 1.0, i.e. all dominant reaction
    pathways for the flux network are computed. For large netorks the
    number of possible reaction paths can increase rapidly so that it
    becomes prohibitevely expensive to compute all possible reaction
    paths. To prevent this from happening maxiter sets the maximum
    number of reaction pathways that will be computed.

    For large flux networks it might be necessary to decrease fraction
    or to increase maxiter. It is advisable to begin with a small
    value for fraction and monitor the number of pathways returned
    when increasing the value of fraction.

    References
    ----------
    .. [1] P. Metzner, C. Schuette and E. Vanden-Eijnden.
        Transition Path Theory for Markov Jump Processes.
        Multiscale Model Simul 7: 1192-1219 (2009)

    """
    from .pathways import pathways as impl
    if issparse(F):
        return impl(F, A, B, fraction=fraction, maxiter=maxiter)
    return impl(csr_matrix(F), A, B, fraction=fraction, maxiter=maxiter)



from typing import Optional

import numpy as np

from ._cluster_model import ClusterModel
from ..base import Estimator


class BoxDiscretizationModel(ClusterModel):
    r""" Model produced by :class:`BoxDiscretization`. Can be used to discretize and/or one-hot transform data.

    Parameters
    ----------
    cluster_centers : ndarray
        The cluster centers.
    v0 : ndarray
        Lower left vertex of box.
    v1 : ndarray
        Upper right vertex of box.
    n_boxes : int
        Number of boxes.
    """

    def __init__(self, cluster_centers: np.ndarray, v0, v1, n_boxes):
        super().__init__(cluster_centers)
        self.v0 = v0
        self.v1 = v1
        self.n_boxes = n_boxes

    def transform_onehot(self, data, n_jobs=None):
        r"""Transforms data into discrete states with subsequent one-hot encoding.

        Parameters
        ----------
        data : ndarray
            Input data
        n_jobs : int or None, optional, default=None
            Number of jobs.

        Returns
        -------
        one_hot : ndarray
            A (T, n_boxes) shaped array with one-hot encoded data.
        """
        dtraj = self.transform(data, n_jobs=n_jobs)
        traj_onehot = np.zeros((len(data), self.n_clusters))
        traj_onehot[np.arange(len(data)), dtraj] = 1.
        return traj_onehot


class BoxDiscretization(Estimator):
    r"""An n-dimensional box discretization of Euclidean space.

    It spans an n-dimensional grid based on linspaces along each axis which is then used as cluster centers.
    The linspaces are bounded either by the user (attributes :attr:`v0` and :attr:`v1`) or estimated from data.

    Parameters
    ----------
    dim : int
        Dimension of the box discretization.
    n_boxes : int or list of int
        Number of boxes per dimension of - if given as single integer - for all dimensions.
    v0 : array or None, optional, default=None
        Lower left vertex of the box discretization. If not given this is estimated from data.
    v1 : array or None, optional, default=None
        Upper right vertex of the box discretization. If not given this is estimated from data.
    """

    def __init__(self, dim: int, n_boxes, v0=None, v1=None):
        super().__init__()
        if not isinstance(n_boxes, (list, tuple, np.ndarray)):
            if int(n_boxes) == n_boxes:
                n_boxes = [int(n_boxes)] * dim
        if len(n_boxes) != dim:
            raise ValueError(f"Dimension and number of boxes per dimension did not match ({len(n_boxes)} and {dim}).")
        if v0 is not None and len(v0) != dim:
            raise ValueError("Length of v0 did not match dimension.")
        if v1 is not None and len(v1) != dim:
            raise ValueError("Length of v1 did not match dimension.")
        self.dim = dim
        self.n_boxes = n_boxes
        self.v0 = v0
        self.v1 = v1

    def fit(self, data: np.ndarray, **kwargs):
        assert data.shape[1] == self.dim
        if self.v0 is None or self.v1 is None:
            v0 = np.empty((self.dim,), dtype=data.dtype) if self.v0 is None else self.v0
            v1 = np.empty((self.dim,), dtype=data.dtype) if self.v1 is None else self.v1
            for d in range(self.dim):
                if self.v0 is None:
                    v0[d] = np.min(data[:, d])
                if self.v1 is None:
                    v1[d] = np.max(data[:, d])
        else:
            v0 = self.v0
            v1 = self.v1
        linspaces = [np.linspace(v0[d], v1[d], num=self.n_boxes[d], endpoint=True) for d in range(self.dim)]
        mesh = np.vstack(np.meshgrid(*tuple(linspaces))).reshape(self.dim, -1).T
        self._model = BoxDiscretizationModel(mesh, v0, v1, self.n_boxes)
        return self

    def fetch_model(self) -> Optional[BoxDiscretizationModel]:
        r""" Yields the estimated model.

        Returns
        -------
        model : BoxDiscretizationModel or None
            The model.
        """
        return super().fetch_model()



import functools
from typing import List

import numpy as _np

__author__ = 'noe, marscher, clonker'


@functools.lru_cache(maxsize=1)
def _load_double_well_discrete():
    import os
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data', 'double_well_discrete.npz')
    with _np.load(filename) as datafile:
        dtraj = datafile['dtraj']
        transition_matrix = datafile['P']
    # avoid side effects, since we are caching these arrays!
    dtraj.flags.writeable = False
    transition_matrix.flags.writeable = False

    from deeptime.markov.msm import MarkovStateModel
    msm = MarkovStateModel(transition_matrix)
    return dtraj, msm


class DoubleWellDiscrete:
    r""" MCMC process in a symmetric double well potential, spatially discretized to 100 bins.

    Encapsulates discrete trajectories and
    markov state model (see :class:`MarkovStateModel <deeptime.markov.msm.MarkovStateModel>`) with exact
    transition matrix.
    """

    def __init__(self):
        dtraj, msm = _load_double_well_discrete()
        self._dtraj = dtraj
        self._analytic_msm = msm

    @property
    def dtraj(self):
        """ 100K frames trajectory at timestep 10, 100 microstates (not all are populated). """
        return self._dtraj

    @property
    def dtraj_n2good(self):
        """ 100K frames trajectory at timestep 10, good 2-state discretization (at transition state). """
        return self.dtraj_n([50])

    @property
    def dtraj_n2bad(self):
        """ 100K frames trajectory at timestep 10, bad 2-state discretization (off transition state). """
        return self.dtraj_n([40])

    def dtraj_n2(self, divide):
        """ 100K frames trajectory at timestep 10, arbitrary 2-state discretization. """
        return self.dtraj_n([divide])

    @property
    def dtraj_n6good(self):
        """ 100K frames trajectory at timestep 10, good 6-state discretization. """
        return self.dtraj_n([40, 45, 50, 55, 60])

    def dtraj_n(self, divides):
        r""" 100K frames trajectory at timestep 10, arbitrary n-state discretization.

        Parameters
        ----------
        divides : (n, dtype=int) ndarray
            The state boundaries.

        Returns
        -------
        dtraj : (T,) ndarray
            Discrete trajectory with :code:`len(divides)` states.
        """
        disc = _np.zeros(100, dtype=int)
        divides = _np.concatenate([divides, [100]])
        for i in range(len(divides) - 1):
            disc[divides[i]:divides[i + 1]] = i + 1
        return disc[self.dtraj]

    @property
    def transition_matrix(self):
        """ Exact transition matrix used to generate the data """
        return self.analytic_msm.transition_matrix

    @property
    def analytic_msm(self):
        """ Returns a :class:`MarkovStateModel <deeptime.markov.msm.MarkovStateModel>` instance with
        the exact transition matrix. """
        return self._analytic_msm

    def simulate_trajectory(self, n_steps, start=None, stop=None, dt=1) -> _np.ndarray:
        """ Generates a discrete trajectory of length less or equal n_steps.

        Parameters
        ----------
        n_steps : int
            maximum number of steps to simulate
        start : int, optional, default=None
            Starting state. If None is given, it is sampled from the stationary distribution.
        stop : int, optional, default=None
            Stopping state. If not None and encountered, stops the simulation. This can lead to fewer
            than n_steps steps.
        dt : int, optional, default=1
            Time step to apply when simulating the trajectory.

        Returns
        -------
        dtraj : (T, 1) ndarray
            A discrete trajectory.
        """
        return self.analytic_msm.simulate(n_steps, start=start, stop=stop, dt=dt)

    def simulate_trajectories(self, n_trajectories: int, n_steps: int,
                              start=None, stop=None, dt=1) -> List[_np.ndarray]:
        """
        Simulates :code:`n_trajectories` discrete trajectories. For a more detailed description of the arguments, see
        :meth:`simulate_trajectory`.
        """
        return [self.simulate_trajectory(n_steps, start=start, stop=stop, dt=dt) for _ in range(n_trajectories)]



r"""
.. currentmodule: deeptime.decomposition

===============================================================================
Estimators
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    VAMP
    TICA
    DMD
    EDMD
    KernelEDMD
    KernelCCA
    KVAD

===============================================================================
Deep estimators
===============================================================================

Note that usage of these estimators requires a working installation of `PyTorch <https://pytorch.org/>`__.

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    deep.VAMPNet
    deep.TAE
    deep.TVAE

===============================================================================
Models
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    TransferOperatorModel
    CovarianceKoopmanModel
    DMDModel
    EDMDModel

    KernelEDMDModel
    KernelCCAModel
    KVADModel

    deep.VAMPNetModel
    deep.TAEModel
    deep.TVAEModel


===============================================================================
Utils
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    vamp_score
    vamp_score_data
    vamp_score_cv

    cvsplit_trajs

    deep.TVAEEncoder
    deep.koopman_matrix
    deep.sym_inverse
    deep.covariances
    deep.vamp_score
    deep.vampnet_loss
    deep.kvad_score
"""

from ._score import vamp_score, vamp_score_data, vamp_score_cv, cvsplit_trajs
from ._tica import TICA
from ._vamp import VAMP
from ._koopman import TransferOperatorModel, CovarianceKoopmanModel
from ._dmd import DMD, DMDModel, EDMD, EDMDModel, KernelEDMD, KernelEDMDModel
from ._cca import KernelCCA, KernelCCAModel
from ._kvad import KVAD, KVADModel

from ..util.platform import module_available
if module_available("torch"):
    from . import deep
del module_available



import torch
from . import _vampnet as vnet
from ...kernels import Kernel, GaussianKernel, is_torch_kernel


def whiten(data, epsilon=1e-6, mode='clamp'):
    data_meanfree = data - data.mean(dim=0, keepdim=True)
    cov = 1 / (data.shape[0] - 1) * (data_meanfree.t() @ data_meanfree)
    cov_sqrt_inv = vnet.sym_inverse(cov, epsilon=epsilon, mode=mode, return_sqrt=True)
    return data_meanfree @ cov_sqrt_inv


def gramian(y, kernel):
    with torch.no_grad():
        if is_torch_kernel(kernel):
            gramian = kernel.gram(y)
        else:
            gramian_np = kernel.gram(y.cpu().numpy())
            gramian = torch.as_tensor(gramian_np, dtype=y.dtype, device=y.device)
    return gramian


def kvad_score(chi_x, y, kernel: Kernel = GaussianKernel(1.), epsilon=1e-6, mode='regularize'):
    G = gramian(y, kernel)

    N = y.shape[0]
    chi_x_whitened = whiten(chi_x, epsilon=epsilon, mode=mode)
    x_g_x = torch.chain_matmul(chi_x_whitened.t(), G, chi_x_whitened)

    evals_sum = torch.trace(x_g_x / (N * N))

    return evals_sum + torch.mean(G)



import math
from typing import Union, Callable, List, Any, Iterable

import numpy as np

from deeptime.util.types import ensure_array


def confidence_interval(data, conf=0.95):
    r""" Computes element-wise confidence intervals from a sample of ndarrays

    Given a sample of arbitrarily shaped ndarrays, computes element-wise
    confidence intervals

    Parameters
    ----------
    data : array-like of dimension 1 and 2
        array of numbers or arrays. The first index is used as the sample
        index, the remaining indexes are specific to the array of interest
    conf : float, optional, default = 0.95
        confidence interval

    Return
    ------
    lower : ndarray(shape)
        element-wise lower bounds
    upper : ndarray(shape)
        element-wise upper bounds

    """
    if conf < 0 or conf > 1:
        raise ValueError(f'Not a meaningful confidence level: {conf}')

    data = ensure_array(data)

    def _confidence_interval_1d(x):
        """
        Computes the mean and alpha-confidence interval of the given sample set

        Parameters
        ----------
        x : ndarray
            a 1D-array of samples

        Returns
        -------
        (m, l, r) : m is the mean of the data, and (l, r) are the m-alpha/2
            and m+alpha/2 confidence interval boundaries.
        """
        assert x.ndim == 1, x.ndim

        if np.any(np.isnan(x)):
            return np.nan, np.nan, np.nan

        d_min, d_max = np.min(x), np.max(x)

        if np.isclose(d_min, d_max):
            return d_min, d_min, d_max

        m = np.mean(x)
        x = np.sort(x)

        # index of the mean
        im = np.searchsorted(x, m)
        if im == 0 or im == len(x) or (np.isinf(m - x[im - 1]) and np.isinf(x[im] - x[im - 1])):
            pm = im
        else:
            pm = (im - 1) + (m - x[im - 1]) / (x[im] - x[im - 1])
        # left interval boundary
        pl = pm - conf * pm
        left_boundary = _compute_interval_boundary(pl, x)
        # right interval boundary
        pr = pm + conf * (len(x) - im)
        right_boundary = _compute_interval_boundary(pr, x)
        return m, left_boundary, right_boundary

    def _compute_interval_boundary(p, x):
        i1 = max(0, int(math.floor(p)))
        i2 = min(len(x) - 1, int(math.ceil(p)))
        if np.isclose(x[i1], x[i2]):  # catch infs
            return x[i1]
        else:
            return x[i1] + (p - i1) * (x[i2] - x[i1])

    if conf == 1.:
        return np.min(data, axis=0), np.max(data, axis=0)

    if data.ndim == 1:
        mean, lower, upper = _confidence_interval_1d(data)
        return lower, upper
    else:
        lower = np.zeros_like(data[0])
        upper = np.zeros_like(data[0])

        # compute interval for each column

        def _column(arr, indexes):
            """ Returns a column with given indexes from a deep array

            For example, if the array is a matrix and indexes is a single int, will
            return arr[:,indexes]. If the array is an order 3 tensor and indexes is a
            pair of ints, will return arr[:,indexes[0],indexes[1]], etc.

            """
            if arr.ndim == 2 and isinstance(indexes, (int, tuple)):
                if isinstance(indexes, tuple):
                    indexes = indexes[0]
                return arr[:, indexes]
            elif arr.ndim == 3 and len(indexes) == 2:
                return arr[:, indexes[0], indexes[1]]
            else:
                raise NotImplementedError('Only supporting arrays of dimension 2 and 3 as yet.')

        for i in np.ndindex(data[0].shape):
            col = _column(data, i)
            mean, lower[i], upper[i] = _confidence_interval_1d(col)

        return lower, upper


def call_member(obj, f: Union[str, Callable], *args, **kwargs):
    """ Calls the specified method, property or attribute of the given object

    Parameters
    ----------
    obj : object
        The object that will be used
    f : str or function
        Name of or reference to method, property or attribute
    *args : list
        list of arguments to pass to f during evaluation
    ** kwargs: dict
        keyword arguments to pass to f during evaluation
    """
    import inspect
    # get function name
    if not isinstance(f, str):
        fname = f.__func__.__name__
    else:
        fname = f
    # get the method ref
    method = getattr(obj, fname)
    # handle cases
    if inspect.ismethod(method):
        return method(*args, **kwargs)

    # attribute or property
    return method


def evaluate_samples(samples: Iterable[Any], quantity: str, delimiter: str = '/', *args, **kwargs) -> List[Any]:
    r"""Evaluate a quantity (like a property, function, attribute) of an iterable of objects and return the result.

    Parameters
    ----------
    samples : list of object
        The samples which contain sought after quantities.
    quantity : str, optional, default=None
        Name of attribute, which will be evaluated on samples. If None, no quantity is evaluated and the samples
        are assumed to already be the quantity that is to be evaluated.
    delimiter : str, optional, default='/'
        Separator to call members of members.
    *args
        pass through
    **kwargs
        pass through

    Returns
    -------
    result : list of any or ndarray
        The collected data, if it can be converted to numpy array then numpy array.
    """
    if quantity is not None and delimiter in quantity:
        qs = quantity.split(delimiter)
        quantity = qs[-1]
        for q in qs[:-1]:
            samples = [call_member(s, q) for s in samples]
    if quantity is not None:
        samples = [call_member(s, quantity, *args, **kwargs) for s in samples]
    try:
        samples = np.asfarray(samples)
    except:
        pass
    return samples


class QuantityStatistics:
    """ Container for statistical quantities computed on samples.

    Parameters
    ----------
    samples: list of ndarrays
        the samples
    store_samples: bool, default=False
        whether to store the samples (array).
    """

    @property
    def R(self):
        r"""Element-wise upper bounds.

        :type: ndarray
        """
        return self._R

    @property
    def L(self):
        r""" Element-wise lower bounds.

        :type: ndarray
        """
        return self._L

    @property
    def std(self):
        r""" Standard deviation along axis=0.

        :type: (n,) ndarray
        """
        return self._std

    @property
    def mean(self):
        r""" Mean along axis=0

        :type: (n,) ndarray
        """
        return self._mean

    @staticmethod
    def gather(samples, quantity=None, store_samples=False, delimiter='/', confidence: float = 0.95, *args, **kwargs):
        r"""Obtain statistics about a sampled quantity. Can also be a chained call, separated by the delimiter.

        Parameters
        ----------
        samples : list of object
            The samples which contain sought after quantities.
        quantity : str, optional, default=None
            Name of attribute, which will be evaluated on samples. If None, no quantity is evaluated and the samples
            are assumed to already be the quantity that is to be evaluated.
        store_samples : bool, optional, default=False
            Whether to store the samples (array).
        delimiter : str, optional, default='/'
            Separator to call members of members.
        confidence : float, optional, default=0.95
            Confidence parameter for the confidence intervals.
        *args
            pass through
        **kwargs
            pass through

        Returns
        -------
        statistics : QuantityStatistics
            The collected statistics.
        """
        if quantity is not None:
            samples = evaluate_samples(samples, quantity=quantity, delimiter=delimiter, *args, **kwargs)
        return QuantityStatistics(samples, quantity=quantity, store_samples=store_samples, confidence=confidence)

    def __init__(self, samples: List[np.ndarray], quantity, confidence=0.95, store_samples=False):
        super().__init__()
        self.quantity = quantity
        samples = np.array(samples)
        if store_samples:
            self.samples = samples
        else:
            self.samples = np.empty(0)
        self._mean = samples.mean(axis=0)
        self._std = samples.std(axis=0)
        self._L, self._R = confidence_interval(samples, conf=confidence)

    def __str__(self):
        return "QuantityStatistics(mean={}, std={})".format(self.mean, self.std)


def _maxlength(X):
    """ Returns the maximum length of signal trajectories X """
    return np.fromiter((map(lambda x: len(x), X)), dtype=int).max()


def statistical_inefficiency(X, truncate_acf=True, mact=1.0):
    r""" Estimates the statistical inefficiency from univariate time series X

    The statistical inefficiency [1]_ is a measure of the correlatedness of samples in a signal.
    Given a signal :math:`{x_t}` with :math:`N` samples and statistical inefficiency :math:`I \in (0,1]`, there are
    only :math:`I \cdot N` effective or uncorrelated samples in the signal. This means that :math:`I \cdot N` should
    be used in order to compute statistical uncertainties. See [2]_ for a review.

    The statistical inefficiency is computed as :math:`I = (2 \tau)^{-1}` using the damped autocorrelation time

    .. math:
        \tau = \frac{1}{2}+\sum_{K=1}^{N} A(k) \left(1-\frac{k}{N}\right)

    where

    .. math:
        A(k) = \frac{\langle x_t x_{t+k} \rangle_t - \langle x^2 \rangle_t}{\mathrm{var}(x)}

    is the autocorrelation function of the signal :math:`{x_t}`, which is computed either for a single or multiple
    trajectories.

    Parameters
    ----------
    X : float array or list of float arrays
        Univariate time series (single or multiple trajectories)
    truncate_acf : bool, optional, default=True
        When the normalized autocorrelation function passes through 0, it is truncated in order to avoid integrating
        random noise

    References
    ----------
    .. [1] Anderson, T. W.: The Statistical Analysis of Time Series (Wiley, New York, 1971)

    .. [2] Janke, W: Statistical Analysis of Simulations: Data Correlations and Error Estimation
        Quantum Simulations of Complex Many-Body Systems: From Theory to Algorithms, Lecture Notes,
        J. Grotendorst, D. Marx, A. Muramatsu (Eds.), John von Neumann Institute for Computing, Juelich
        NIC Series 10, pp. 423-445, 2002.

    """
    # check input
    assert np.ndim(X[0]) == 1, 'Data must be 1-dimensional'
    N = _maxlength(X)  # max length
    # mean-free data
    xflat = np.concatenate(X)
    Xmean = np.mean(xflat)
    X0 = [x - Xmean for x in X]
    # moments
    x2m = np.mean(xflat ** 2)
    del xflat, Xmean
    # integrate damped autocorrelation
    corrsum = 0.0
    for lag in range(N):
        acf = 0.0
        n = 0
        # cache partial sums
        for x in X0:
            Nx = len(x)  # length of this trajectory
            if Nx > lag:  # only use trajectories that are long enough
                prod = x[:Nx - lag] * x[lag:]
                acf += np.sum(prod)
                n += Nx - lag
        acf /= float(n)
        if acf <= 0 and truncate_acf:  # zero autocorrelation. Exit
            break
        elif lag > 0:  # start integrating at lag 1 (effect of lag 0 is contained in the 0.5 below
            corrsum += acf * (1.0 - (float(lag) / float(N)))
    # compute damped correlation time
    corrtime = 0.5 + mact * corrsum / x2m
    # return statistical inefficiency
    return 1.0 / (2 * corrtime)



"""
@author: paul, marscher, wu, noe, clonker
"""

from collections import namedtuple
from numbers import Integral
from typing import Optional, Union, Callable

import numpy as np

from ._koopman import CovarianceKoopmanModel, KoopmanChapmanKolmogorovValidator
from ..base import EstimatorTransformer
from ..basis import Identity
from ..covariance import Covariance, CovarianceModel
from ..numeric import spd_inv_split
from ..util.types import to_dataset


class VAMP(EstimatorTransformer):
    r"""Variational approach for Markov processes (VAMP).

    The implementation is based on :footcite:`wu2020variational`, :footcite:`noe2015kinetic`.

    Parameters
    ----------
    lagtime : int or None, optional, default=None
        The lagtime under which covariances are estimated. This is only relevant when estimating from data, in case
        covariances are provided this should either be None or exactly the value that was used to estimate
        said covariances.
    dim : int, optional, default=None
        Number of dimensions to keep:

        * if dim is not set (None) all available ranks are kept:
          :code:`n_components == min(n_samples, n_uncorrelated_features)`
        * if dim is an integer >= 1, this number specifies the number
          of dimensions to keep.
    var_cutoff : float, optional, default=None
        Determines the number of output dimensions by including dimensions until their cumulative kinetic variance
        exceeds the fraction subspace variance. var_cutoff=1.0 means all numerically available dimensions
        (see epsilon) will be used, unless set by dim. Setting var_cutoff smaller than 1.0 is exclusive with dim.
    scaling : str, optional, default=None
        Scaling to be applied to the VAMP order parameters upon transformation

        * None: no scaling will be applied, variance of the order parameters is 1
        * 'kinetic_map' or 'km': order parameters are scaled by singular value.
          Only the left singular functions induce a kinetic map wrt the
          conventional forward propagator. The right singular functions induce
          a kinetic map wrt the backward propagator.
    epsilon : float, optional, default=1e-6
        Eigenvalue cutoff. Eigenvalues of :math:`C_{00}` and :math:`C_{11}`
        with norms <= epsilon will be cut off. The remaining number of
        eigenvalues together with the value of `dim` define the size of the output.
    observable_transform : callable, optional, default=Identity
        A feature transformation on the raw data which is used to estimate the model.

    See Also
    --------
    CovarianceKoopmanModel : type of model produced by this estimator

    Notes
    -----
    VAMP is a method for dimensionality reduction of Markov processes.

    The Koopman operator :math:`\mathcal{K}` is an integral operator
    that describes conditional future expectation values. Let
    :math:`p(\mathbf{x},\,\mathbf{y})` be the conditional probability
    density of visiting an infinitesimal phase space volume around
    point :math:`\mathbf{y}` at time :math:`t+\tau` given that the phase
    space point :math:`\mathbf{x}` was visited at the earlier time
    :math:`t`. Then the action of the Koopman operator on a function
    :math:`f` can be written as follows:

    .. math::

      \mathcal{K}f=\int p(\mathbf{x},\,\mathbf{y})f(\mathbf{y})\,\mathrm{dy}=\mathbb{E}\left[f(\mathbf{x}_{t+\tau}\mid\mathbf{x}_{t}=\mathbf{x})\right]

    The Koopman operator is defined without any reference to an
    equilibrium distribution. Therefore it is well-defined in
    situations where the dynamics is irreversible or/and non-stationary
    such that no equilibrium distribution exists.

    If we approximate :math:`f` by a linear superposition of ansatz
    functions :math:`\boldsymbol{\chi}` of the conformational
    degrees of freedom (features), the operator :math:`\mathcal{K}`
    can be approximated by a (finite-dimensional) matrix :math:`\mathbf{K}`.

    The approximation is computed as follows: From the time-dependent
    input features :math:`\boldsymbol{\chi}(t)`, we compute the mean
    :math:`\boldsymbol{\mu}_{0}` (:math:`\boldsymbol{\mu}_{1}`) from
    all data excluding the last (first) :math:`\tau` steps of every
    trajectory as follows:

    .. math::

      \boldsymbol{\mu}_{0}	:=\frac{1}{T-\tau}\sum_{t=0}^{T-\tau}\boldsymbol{\chi}(t)

      \boldsymbol{\mu}_{1}	:=\frac{1}{T-\tau}\sum_{t=\tau}^{T}\boldsymbol{\chi}(t)

    Next, we compute the instantaneous covariance matrices
    :math:`\mathbf{C}_{00}` and :math:`\mathbf{C}_{11}` and the
    time-lagged covariance matrix :math:`\mathbf{C}_{01}` as follows:

    .. math::

        \begin{aligned}
      \mathbf{C}_{00}&:=\frac{1}{T-\tau}\sum_{t=0}^{T-\tau}\left[\boldsymbol{\chi}(t)-\boldsymbol{\mu}_{0}\right]\left[\boldsymbol{\chi}(t)-\boldsymbol{\mu}_{0}\right]\\
      \mathbf{C}_{11}&:=\frac{1}{T-\tau}\sum_{t=\tau}^{T}\left[\boldsymbol{\chi}(t)-\boldsymbol{\mu}_{1}\right]\left[\boldsymbol{\chi}(t)-\boldsymbol{\mu}_{1}\right]\\
      \mathbf{C}_{01}&:=\frac{1}{T-\tau}\sum_{t=0}^{T-\tau}\left[\boldsymbol{\chi}(t)-\boldsymbol{\mu}_{0}\right]\left[\boldsymbol{\chi}(t+\tau)-\boldsymbol{\mu}_{1}\right]
        \end{aligned}

    The Koopman matrix is then computed as follows:

    .. math::

      \mathbf{K}=\mathbf{C}_{00}^{-1}\mathbf{C}_{01}

    It can be shown :footcite:`wu2020variational` that the leading singular functions of the
    half-weighted Koopman matrix

    .. math::

      \bar{\mathbf{K}}:=\mathbf{C}_{00}^{-\frac{1}{2}}\mathbf{C}_{01}\mathbf{C}_{11}^{-\frac{1}{2}}

    encode the best reduced dynamical model for the time series.

    The singular functions can be computed by first performing the
    singular value decomposition

    .. math::

      \bar{\mathbf{K}}=\mathbf{U}^{\prime}\mathbf{S}\mathbf{V}^{\prime}

    and then mapping the input conformation to the left singular
    functions :math:`\boldsymbol{\psi}` and right singular
    functions :math:`\boldsymbol{\phi}` as follows:

    .. math::

        \begin{aligned}
        \boldsymbol{\psi}(t)&:=\mathbf{U}^{\prime\top}\mathbf{C}_{00}^{-\frac{1}{2}}\left[\boldsymbol{\chi}(t)-\boldsymbol{\mu}_{0}\right]\\
        \boldsymbol{\phi}(t)&:=\mathbf{V}^{\prime\top}\mathbf{C}_{11}^{-\frac{1}{2}}\left[\boldsymbol{\chi}(t)-\boldsymbol{\mu}_{1}\right]
        \end{aligned}


    References
    ----------
    .. footbibliography::
    """

    def __init__(self, lagtime: Optional[int] = None,
                 dim: Optional[int] = None,
                 var_cutoff: Optional[float] = None,
                 scaling: Optional[str] = None,
                 epsilon: float = 1e-6,
                 observable_transform: Callable[[np.ndarray], np.ndarray] = Identity()):
        super(VAMP, self).__init__()
        self.dim = dim
        self.var_cutoff = var_cutoff
        self.scaling = scaling
        self.epsilon = epsilon
        self.lagtime = lagtime
        self.observable_transform = observable_transform
        self._covariance_estimator = None  # internal covariance estimator

    _DiagonalizationResults = namedtuple("DiagonalizationResults", ['rank0', 'rankt', 'singular_values',
                                                                    'left_singular_vecs', 'right_singular_vecs'])

    @staticmethod
    def _decomposition(covariances, epsilon, scaling, dim, var_cutoff) -> _DiagonalizationResults:
        """Performs SVD on covariance matrices and save left, right singular vectors and values in the model."""
        L0 = spd_inv_split(covariances.cov_00, epsilon=epsilon)
        rank0 = L0.shape[1] if L0.ndim == 2 else 1
        Lt = spd_inv_split(covariances.cov_tt, epsilon=epsilon)
        rankt = Lt.shape[1] if Lt.ndim == 2 else 1

        W = np.dot(L0.T, covariances.cov_0t).dot(Lt)
        from scipy.linalg import svd
        A, s, BT = svd(W, compute_uv=True, lapack_driver='gesvd')

        singular_values = s

        m = CovarianceKoopmanModel.effective_output_dimension(rank0, rankt, dim, var_cutoff, singular_values)

        U = np.dot(L0, A[:, :m])
        V = np.dot(Lt, BT[:m, :].T)

        # scale vectors
        if scaling is not None and scaling in ("km", "kinetic_map"):
            U *= s[np.newaxis, 0:m]  # scaled left singular functions induce a kinetic map
            V *= s[np.newaxis, 0:m]  # scaled right singular functions induce a kinetic map wrt. backward propagator

        return VAMP._DiagonalizationResults(
            rank0=rank0, rankt=rankt, singular_values=singular_values, left_singular_vecs=U, right_singular_vecs=V
        )

    @classmethod
    def covariance_estimator(cls, lagtime: int, ncov: Union[int] = float('inf')):
        r""" Yields a properly configured covariance estimator so that its model can be used as input for the vamp
        estimator.

        Parameters
        ----------
        lagtime : int
            Positive integer denoting the time shift which is considered for autocorrelations.
        ncov : int or float('inf'), optional, default=float('inf')
            Limit the memory usage of the algorithm from :footcite:`chan1982updating` to an amount that corresponds
            to ncov additional copies of each correlation matrix.

        Returns
        -------
        estimator : Covariance
            Covariance estimator.
        """
        return Covariance(lagtime=lagtime, compute_c0t=True, compute_ctt=True, remove_data_mean=True, reversible=False,
                          bessels_correction=False, ncov=ncov)

    @staticmethod
    def _to_covariance_model(covariances: Union[Covariance, CovarianceModel]) -> CovarianceModel:
        if isinstance(covariances, Covariance):
            covariances = covariances.fetch_model()
        return covariances

    def partial_fit(self, data):
        r""" Updates the covariance estimates through a new batch of data.

        Parameters
        ----------
        data : tuple(ndarray, ndarray)
            A tuple of ndarrays which have to have same shape and are :math:`X_t` and :math:`X_{t+\tau}`, respectively.
            Here, :math:`\tau` denotes the lagtime.

        Returns
        -------
        self : VAMP
            Reference to self.
        """
        if self._covariance_estimator is None:
            self._covariance_estimator = self.covariance_estimator(lagtime=self.lagtime)
        x, y = to_dataset(data, lagtime=self.lagtime)[:]
        self._covariance_estimator.partial_fit((self.observable_transform(x),
                                                self.observable_transform(y)))
        return self

    def fit_from_covariances(self, covariances: Union[Covariance, CovarianceModel]):
        r"""Fits from existing covariance model (or covariance estimator containing model).

        Parameters
        ----------
        covariances : CovarianceModel or Covariance
            Covariance model containing covariances or Covariance estimator containing a covariance model. The model
            in particular has matrices :math:`C_{00}, C_{0t}, C_{tt}`.

        Returns
        -------
        self : VAMP
            Reference to self.
        """
        self._covariance_estimator = None
        covariances = self._to_covariance_model(covariances)
        self._model = self._decompose(covariances)
        return self

    def fit_from_timeseries(self, data, weights=None):
        r""" Estimates a :class:`CovarianceKoopmanModel` directly from time-series data using the :class:`Covariance`
        estimator. For parameters `dim`, `scaling`, `epsilon`.

        Parameters
        ----------
        data
            Input data, see :meth:`to_dataset <deeptime.util.types.to_dataset>` for options.
        weights
            See the :class:`Covariance <deeptime.covariance.Covariance>` estimator.

        Returns
        -------
        self : VAMP
            Reference to self.
        """
        dataset = to_dataset(data, lagtime=self.lagtime)
        self._covariance_estimator = self.covariance_estimator(lagtime=self.lagtime)
        x, y = dataset[:]
        transformed = (self.observable_transform(x), self.observable_transform(y))
        covariances = self._covariance_estimator.partial_fit(transformed, weights=weights)\
            .fetch_model()
        return self.fit_from_covariances(covariances)

    @property
    def epsilon(self):
        r""" Eigenvalue cutoff.

        :getter: Yields current eigenvalue cutoff.
        :setter: Sets new eigenvalue cutoff.
        :type: float
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value):
        self._epsilon = value

    @property
    def lagtime(self) -> Optional[int]:
        r""" The lagtime under which covariances are estimated. Can be `None` in case covariances are provided
        directly instead of estimating them inside this estimator.

        :getter: Yields the current lagtime.
        :setter: Sets a new lagtime, must be positive.
        :type: int or None
        """
        return self._lagtime

    @lagtime.setter
    def lagtime(self, value: Optional[int]):
        if value is not None and value <= 0:
            raise ValueError(f"Lagtime needs to be strictly positive but was {value}")
        self._lagtime = value

    @property
    def scaling(self) -> Optional[str]:
        r""" Scaling parameter to be applied to order parameters upon transformation. Can be one of None, 'kinetic_map',
        or 'km'.

        :getter: Yields currently configured scaling parameter.
        :setter: Sets a new scaling parameter (None, 'kinetic_map', or 'km')
        :type: str or None
        """
        return self._scaling

    @scaling.setter
    def scaling(self, value: Optional[str]):
        valid = (None, 'km', 'kinetic_map')
        if value not in valid:
            raise ValueError("Invalid scaling parameter \"{}\", can only be one of {}".format(value, valid))
        self._scaling = value

    @property
    def var_cutoff(self) -> Optional[float]:
        r""" Variational cutoff which can be used to further restrict the dimension. This takes precedence over the
        :meth:`dim` property.

        :getter: yields the currently set variation cutoff
        :setter: sets a new cutoff
        :type: float or None
        """
        return self._var_cutoff

    @var_cutoff.setter
    def var_cutoff(self, value: Optional[float]):
        if value is not None and (value <= 0. or float(value) > 1.0):
            raise ValueError("VAMP: Invalid var_cutoff parameter, can only be in the interval (0, 1].")
        self._var_cutoff = value

    @property
    def dim(self) -> Optional[int]:
        r""" Dimension attribute. Can either be int or float. In case of

        * :code:`int` it evaluates it as the actual dimension, must be strictly greater 0,
        * :code:`None` all numerically available components are used.

        :getter: yields the dimension
        :setter: sets a new dimension
        :type: int or None
        """
        return self._dim

    @dim.setter
    def dim(self, value: Optional[Union[int, float]]):
        if isinstance(value, Integral):
            if value <= 0:
                # first test against Integral as `isinstance(1, Real)` also evaluates to True
                raise ValueError("VAMP: Invalid dimension parameter, if it is given in terms of the "
                                 "dimension (integer), must be positive.")
        elif value is not None:
            raise ValueError("Invalid type for dimension, got {}".format(value))
        self._dim = value

    def _decompose(self, covariances: CovarianceModel):
        decomposition = self._decomposition(covariances, self.epsilon, self.scaling, self.dim, self.var_cutoff)
        return CovarianceKoopmanModel(
            decomposition.left_singular_vecs, decomposition.singular_values, decomposition.right_singular_vecs,
            rank_0=decomposition.rank0, rank_t=decomposition.rankt, dim=self.dim,
            var_cutoff=self.var_cutoff, cov=covariances, scaling=self.scaling, epsilon=self.epsilon,
            instantaneous_obs=self.observable_transform,
            timelagged_obs=self.observable_transform
        )

    def fit(self, data, *args, **kw):
        r""" Fits a new :class:`CovarianceKoopmanModel` which can be obtained by a
        subsequent call to :meth:`fetch_model`.

        Parameters
        ----------
        data : CovarianceModel or Covariance or timeseries
            Covariance matrices :math:`C_{00}, C_{0t}, C_{tt}` in form of a CovarianceModel instance. If the model
            should be fitted directly from data, please see :meth:`from_data`.
            Optionally, this can also be timeseries data directly, in which case a 'lagtime' must be provided.
        *args
            Optional arguments
        **kw
            Ignored keyword arguments for scikit-learn compatibility.

        Returns
        -------
        self : VAMP
            Reference to self.
        """
        if isinstance(data, (Covariance, CovarianceModel)):
            self.fit_from_covariances(data)
        else:
            self.fit_from_timeseries(data, weights=kw.pop('weights', None))
        return self

    def transform(self, data, propagate=False):
        r""" Projects given timeseries onto dominant singular functions. This method dispatches to
        :meth:`CovarianceKoopmanModel.transform`.

        Parameters
        ----------
        data : (T, n) ndarray
            Input timeseries data.
        propagate : bool, default=False
            Whether to apply the Koopman operator after data was transformed into the whitened feature space.

        Returns
        -------
        Y : (T, m) ndarray
            The projected data.
            If `right` is True, projection will be on the right singular functions. Otherwise, projection will be on
            the left singular functions.
        """
        return self.fetch_model().transform(data, propagate=propagate)

    def fetch_model(self) -> CovarianceKoopmanModel:
        r""" Finalizes current model and yields new :class:`CovarianceKoopmanModel`.

        Returns
        -------
        model : CovarianceKoopmanModel
            The estimated model.
        """
        if self._covariance_estimator is not None:
            # This can only occur when partial_fit was called.
            # A call to fit, fit_from_timeseries, fit_from_covariances ultimately always leads to a call to
            # fit_from_covariances which sets the self._covariance_estimator to None.
            self._model = self._decompose(self._covariance_estimator.fetch_model())
            self._covariance_estimator = None
        return self._model

    def chapman_kolmogorov_validator(self, mlags, test_model: CovarianceKoopmanModel = None,
                                     n_observables=None, observables='phi', statistics='psi'):
        r"""Returns a Chapman-Kolmogorov validator based on this estimator and a test model.

        Parameters
        ----------
        mlags : int or int-array
            Multiple of lagtimes of the test_model to test against.
        test_model : CovarianceKoopmanModel, optional, default=None
            The model that is tested. If not provided, uses this estimator's encapsulated model.
        n_observables : int, optional, default=None
            Limit the number of default observables (and of default statistics) to this number.
            Only used if `observables` are None or `statistics` are None.
        observables : (input_dimension, n_observables) ndarray
            Coefficients that express one or multiple observables in
            the basis of the input features.
        statistics : (input_dimension, n_statistics) ndarray
            Coefficients that express one or multiple statistics in the basis of the input features.

        Returns
        -------
        validator : KoopmanChapmanKolmogorovValidator
            The validator.

        Notes
        -----
        This method computes two sets of time-lagged covariance matrices

        * estimates at higher lag times :

          .. math::

              \left\langle \mathbf{K}(n\tau)g_{i},f_{j}\right\rangle_{\rho_{0}}

          where :math:`\rho_{0}` is the empirical distribution implicitly defined
          by all data points from time steps 0 to T-tau in all trajectories,
          :math:`\mathbf{K}(n\tau)` is a rank-reduced Koopman matrix estimated
          at the lag-time n*tau and g and f are some functions of the data.
          Rank-reduction of the Koopman matrix is controlled by the `dim`
          parameter of :class:`VAMP <deeptime.decomposition.VAMP>`.

        * predictions at higher lag times :

          .. math::

              \left\langle \mathbf{K}^{n}(\tau)g_{i},f_{j}\right\rangle_{\rho_{0}}

          where :math:`\mathbf{K}^{n}` is the n'th power of the rank-reduced
          Koopman matrix contained in self.

        The Champan-Kolmogorov test is to compare the predictions to the estimates.
        """
        test_model = self.fetch_model() if test_model is None else test_model
        assert test_model is not None, "We need a test model via argument or an estimator which was already" \
                                       "fit to data."
        lagtime = self.lagtime
        if n_observables is not None:
            if n_observables > test_model.dim:
                import warnings
                warnings.warn('Selected singular functions as observables but dimension '
                              'is lower than requested number of observables.')
                n_observables = test_model.dim
        else:
            n_observables = test_model.dim

        if isinstance(observables, str) and observables == 'phi':
            observables = test_model.singular_vectors_right[:, :n_observables]
            observables_mean_free = True
        else:
            observables_mean_free = False

        if isinstance(statistics, str) and statistics == 'psi':
            statistics = test_model.singular_vectors_left[:, :n_observables]
            statistics_mean_free = True
        else:
            statistics_mean_free = False
        return VAMPKoopmanCKValidator(test_model, self, lagtime, mlags, observables, statistics,
                                      observables_mean_free, statistics_mean_free)


def _vamp_estimate_model_for_lag(estimator: VAMP, model, data, lagtime):
    est = VAMP(lagtime=lagtime, dim=estimator.dim, var_cutoff=estimator.var_cutoff, scaling=estimator.scaling,
               epsilon=estimator.epsilon, observable_transform=estimator.observable_transform)
    return est.fit(data).fetch_model()


class VAMPKoopmanCKValidator(KoopmanChapmanKolmogorovValidator):

    def fit(self, data, n_jobs=None, progress=None, **kw):
        return super().fit(data, n_jobs, progress, _vamp_estimate_model_for_lag, **kw)



import numpy as np
from deeptime.base import Model, Transformer

from . import _clustering_bindings as _bd, metrics
from ..util.parallel import handle_n_jobs


class ClusterModel(Model, Transformer):
    r"""
    A generic clustering model. Stores the number of cluster centers, the position of cluster centers, the metric
    that was used to compute the cluster centers and whether the estimation converged. It can be used to transform
    data by assigning each frame to its closest cluster center.

    Parameters
    ----------
    cluster_centers : (k, d) ndarray
        The cluster centers, length of the array should match :attr:`n_clusters`.
    metric : str, default='euclidean'
        The metric that was used for estimation, defaults to Euclidean metric.
    converged : bool, optional, default=False
        Whether the estimation converged.

    Examples
    --------
    Let us create an artificial cluster model with three cluster centers in a three-dimensional space. The cluster
    centers are just the canonical basis vectors :math:`c_1 = (1,0,0)^\top`, :math:`c_2 = (0,1,0)^\top`, and
    :math:`c_3 = (0,0,1)^\top`.

    We can transform data with the model. The data are five frames sampled around the third cluster center.

    >>> model = ClusterModel(cluster_centers=np.eye(3))
    >>> data = np.random.normal(loc=[0, 0, 1], scale=0.01, size=(5, 3))
    >>> assignments = model.transform(data)
    >>> print(assignments)
    [2 2 2 2 2]
    """

    def __init__(self, cluster_centers: np.ndarray, metric: str = 'euclidean',
                 converged: bool = False):
        super().__init__()
        if cluster_centers.ndim == 1:
            cluster_centers = cluster_centers[..., None]
        self._cluster_centers = cluster_centers
        self._metric = metric
        self._converged = converged

    @property
    def cluster_centers(self) -> np.ndarray:
        """
        Gets the cluster centers that were estimated for this model.

        Returns
        -------
        np.ndarray
            Array containing estimated cluster centers.
        """
        return self._cluster_centers

    @property
    def n_clusters(self) -> int:
        """
        The number of cluster centers.

        Returns
        -------
        int
            The number of cluster centers.
        """
        return self.cluster_centers.shape[0]

    @property
    def dim(self):
        return self.cluster_centers.shape[1]

    @property
    def metric(self) -> str:
        """
        The metric that was used.

        Returns
        -------
        metric : str
            Name of the metric that was used. The name is related to the implementation via
            the :data:`metric registry <deeptime.clustering.metrics>`.
        """
        return self._metric

    @property
    def converged(self) -> bool:
        """
        Whether the estimation process converged. Per default this is set to False, which can also indicate that
        the model was created manually and does not stem from an Estimator directly.

        Returns
        -------
        converged : bool
            Whether the clustering converged
        """
        return self._converged

    def transform(self, data, n_jobs=None) -> np.ndarray:
        r"""
        For each frame in `data`, yields the index of the closest point in :attr:`cluster_centers`.

        Parameters
        ----------
        data : (T, d) ndarray
            frames
        n_jobs : int, optional, default=None
            number of jobs to use for assignment

        Returns
        -------
        discrete_trajectory : (T, 1) ndarray
            A discrete trajectory where each frame denotes the closest cluster center.
        """
        n_jobs = handle_n_jobs(n_jobs)
        if data.ndim == 1:
            data = data[..., None]
        dtraj = _bd.assign(data, self.cluster_centers, n_jobs, metrics[self.metric]())
        return dtraj



from pathlib import Path

from numpy.distutils.misc_util import Configuration


def configuration(parent_package='', top_path=None):
    config = Configuration('mle', parent_package, top_path)

    config.add_extension('newton.objective_sparse_ops',
                         sources=['newton/objective_sparse_ops.cpp'],
                         include_dirs=[Path(top_path) / 'deeptime' / 'src' / 'include'],
                         language='c++',
                         extra_compile_args=['-fvisibility=hidden'])

    config.add_subpackage('newton')
    return config



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('hmm', parent_package, top_path)

    config.add_subpackage("init")

    config.add_extension('_hmm_bindings',
                         sources=['_bindings/src/hmm_module.cpp'],
                         include_dirs=['_bindings/include'],
                         language='c++',
                         extra_compile_args=['-fvisibility=hidden']
                         )

    return config



import numpy as np
cimport scipy.linalg.cython_lapack as scc

def eig_qr(A):
    """ Compute eigenvalues and eigenvectors of symmetric matrix A using symmetric tridiagonal QR-algorithm
     with implicit shifts. The matrix is first transformed to tridiagonal shape using lapack's dsytrd routine.
     Then, the tridiagonal QR-iteration is performed using lapack's dsteqr routine.

     Parameters:
     -----------
     A, ndarray (N, N):
        symmetric matrix.

     Returns:
     --------
     D, ndarray(N,)
        array of eigenvalues of A
     B, ndarray(N, N)
        array of eigenvectors of A.
    """

    # handle 1x1 case
    if np.size(A) == 1:  # size can handle 1x1 arrays and numbers
        return A*np.ones(1), np.ones((1, 1))

    # Definitions:
    array_args = dict(dtype=np.float64, order='F')
    cdef double[:,:] B = np.require(A, dtype=np.float64, requirements=('F', 'A'))
    cdef int n = A.shape[0], lda = A.shape[0], info, lwork = -1
    cdef char uplo = b"U"
    cdef double[:] D = np.zeros(n, **array_args)
    cdef double[:] E = np.zeros(n-1, **array_args)
    cdef double[:] Tau = np.zeros(n-1, **array_args)
    cdef double WorkFake # LAPACK writes back the optimal block size here, when lwork is -1.

    # Transform to tridiagonal shape:
    scc.dsytrd(&uplo, &n, &B[0, 0], &lda, &D[0], &E[0], &Tau[0], &WorkFake, &lwork, &info)
    assert info == 0, info
    lwork = <int>WorkFake
    cdef double[:] Work2 = np.zeros(lwork, **array_args)
    scc.dsytrd(&uplo, &n, &B[0, 0], &lda, &D[0], &E[0], &Tau[0], &Work2[0], &lwork, &info)
    assert info == 0, info
    del Work2

    # Extract transformation to tridiagonal shape:
    lwork = -1
    scc.dorgtr(&uplo, &n, &B[0, 0], &lda, &Tau[0], &WorkFake, &lwork, &info)
    assert info == 0, info
    lwork = <int>WorkFake
    cdef double[:] Work3 = np.zeros(lwork, **array_args)
    scc.dorgtr(&uplo, &n, &B[0, 0], &lda, &Tau[0], &Work3[0], &lwork, &info)
    assert info == 0, info
    del Tau, Work3

    # Run QR-iteration.
    cdef double[:] Work4 = np.zeros(max(1, 2*n - 2), **array_args)
    cdef char compz = b"V"
    scc.dsteqr(&compz, &n, &D[0], &E[0], &B[0, 0], &n, &Work4[0], &info)
    assert info == 0, info

    return np.asarray(D), np.asarray(B)



r"""This module provides functions for the computation of stationary
vectors of stochastic matrices

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
.. moduleauthor:: M. Hoffmann
"""
from typing import Optional

import numpy as np
import scipy.sparse as sparse
import scipy.linalg as la
import scipy.sparse.linalg as sla


class _LUSolve:

    def __init__(self, mat):
        self._sparse = sparse.issparse(mat)
        if sparse.issparse(mat):
            self._solve = sla.factorized(mat)
        else:
            self._lu = la.lu_factor(mat)

    def __call__(self, vec):
        if self._sparse:
            return self._solve(vec)
        else:
            return la.lu_solve(self._lu, vec)


def backward_iteration(A, mu, x0, tol=1e-14, maxiter=100):
    r"""Find eigenvector to approximate eigenvalue via backward iteration.

    Parameters
    ----------
    A : (N, N) ndarray or sparse matrix
        Matrix for which eigenvector is desired
    mu : float
        Approximate eigenvalue for desired eigenvector
    x0 : (N, ) ndarray
        Initial guess for eigenvector
    tol : float
        Tolerance parameter for termination of iteration
    maxiter : int
        Maximum number of iterations.

    Returns
    -------
    x : (N, ) ndarray
        Eigenvector to approximate eigenvalue mu

    """
    if sparse.issparse(A):
        T = A - mu * sparse.eye(A.shape[0], A.shape[0])
        T = T.tocsc()
    else:
        T = A - mu * np.eye(A.shape[0])
    lu = _LUSolve(T)  # LU-factor of T

    # Starting iterate with ||y_0||=1
    r0 = 1.0 / np.linalg.norm(x0)
    y0 = x0 * r0
    # Local variables for inverse iteration
    y = 1.0 * y0
    r = 1.0 * r0
    for i in range(maxiter):
        x = lu(y)
        r = 1.0 / np.linalg.norm(x)
        y = x * r
        if r <= tol:
            return y
    msg = "Failed to converge after %d iterations, residuum is %e" % (maxiter, r)
    raise RuntimeError(msg)


def stationary_distribution_from_backward_iteration(P, eps=1e-15):
    r"""Fast computation of the stationary vector using backward
    iteration.

    Parameters
    ----------
    P : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    eps : float (optional)
        Perturbation parameter for the true eigenvalue.

    Returns
    -------
    pi : (M,) ndarray
        Stationary vector

    """
    A = P.transpose()
    mu = 1.0 - eps
    x0 = np.ones(P.shape[0])
    y = backward_iteration(A, mu, x0)
    pi = y / y.sum()
    return pi


def stationary_distribution_from_eigenvector(T, ncv=None):
    r"""Compute stationary distribution of stochastic matrix T.

    The stationary distribution is the left eigenvector corresponding to the
    non-degenerate eigenvalue :math: `\lambda=1`.

    Input:
    ------
    T : numpy array, shape(d,d)
        Transition matrix (stochastic matrix).
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k

    Returns
    -------
    mu : numpy array, shape(d,)
        Vector of stationary probabilities.

    """
    if sparse.issparse(T):
        vals, vecs = sla.eigs(T.transpose(), k=1, which='LR', ncv=ncv)
    else:
        vals, vecs = la.eig(T, left=True, right=False)

    # Sorted eigenvalues and left and right eigenvectors.
    perm = np.argsort(vals)[::-1]
    vecs = vecs[:, perm]

    # Make sure that stationary distribution is non-negative and l1-normalized
    nu = np.abs(vecs[:, 0])
    mu = nu / np.sum(nu)
    return mu


def stationary_distribution(T, ncv: Optional[int] = None, mode: str = 'fallback', check_inputs: bool = True):
    r"""Compute stationary distribution of stochastic matrix T.

    Input:
    ------
    T : (M, M) ndarray or scipy.sparse matrix
        Transition matrix
    ncv : int (optional)
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k. Only relevant for sparse matrices and if backward iteration is unsuccessful.
    mode : str, optional, default='fallback'
        Determines whether the method first tries backward iteration and then eigenvector estimation (`'fallback'`) or
        it uses backward iteration only (`'backward'`) or it uses eigenvector estimation only (`'eigenvector'`).
    check_inputs : bool, optional, default=True
        Whether to check for connectivity and if it is a transition matrix.

    Returns
    -------
    mu : (M,) ndarray
        Vector of stationary probabilities.

    Notes
    -----
    The stationary distribution :math:`\mu` is the left eigenvector
    corresponding to the non-degenerate eigenvalue :math:`\lambda=1`,

    .. math:: \mu^T T =\mu^T.

    Examples
    --------

    >>> import numpy as np
    >>> from deeptime.markov.tools.analysis import stationary_distribution

    >>> T = np.array([[0.9, 0.1, 0.0], [0.4, 0.2, 0.4], [0.0, 0.1, 0.9]])
    >>> mu = stationary_distribution(T)
    >>> mu
    array([0.44444444, 0.11111111, 0.44444444])
    """
    if check_inputs:
        from ._assessment import is_connected, is_transition_matrix
        if not is_transition_matrix(T, tol=1e-12):
            raise ValueError("Input matrix is not a transition matrix. "
                             "Cannot compute stationary distribution")
        if not is_connected(T, directed=False):
            raise ValueError("Input matrix is not weakly connected. "
                             "Therefore it has no unique stationary "
                             "distribution. Separate disconnected components "
                             "and handle them separately")

    assert mode in stationary_distribution.valid_modes, f"Mode must be one of {stationary_distribution.valid_modes}."
    use_backward, use_eigenvector = True, True
    if mode == 'backward':
        use_eigenvector = False
    if mode == 'eigenvector':
        use_backward = False
    assert use_backward or use_eigenvector

    mu = None
    if use_backward:
        try:
            mu = stationary_distribution_from_backward_iteration(T)
        except RuntimeError:
            if not use_eigenvector:
                raise  # we re-raise
            else:
                pass  # try next

    if use_eigenvector and (mu is None or np.any(mu < 0)):
        # numerical problem, fall back to more robust algorithm.
        mu = stationary_distribution_from_eigenvector(T, ncv=ncv)
        if np.any(mu < 0):  # still? Then set to 0 and renormalize
            mu = np.maximum(mu, 0.0)
            mu /= mu.sum()

    return mu


stationary_distribution.valid_modes = ('fallback', 'backward', 'eigenvector')



import numpy as np


class SamplerRev:
    def __init__(self, C, P0=None, seed: int = -1):
        from deeptime.markov.tools.estimation import transition_matrix as tmatrix
        from deeptime.markov.tools.analysis import stationary_distribution
        from .._mle_bindings import RevSampler32, RevSampler64, RevSampler128

        if C.dtype not in (np.float32, np.float64, np.longdouble):
            dtype = np.float64
        else:
            dtype = C.dtype

        self.C = C.astype(dtype)

        """Set up initial state of the chain"""
        if P0 is None:
            # only do a few iterations to get close to the MLE and suppress not converged warning
            P0 = tmatrix(self.C, reversible=True, maxiter=100, warn_not_converged=False)
            assert P0.dtype == self.C.dtype
        else:
            P0 = P0.astype(self.C.dtype)
        pi0 = stationary_distribution(P0).astype(self.C.dtype)
        V0 = pi0[:, np.newaxis] * P0

        self.V = V0
        self.c = self.C.sum(axis=1)

        """Check for valid input"""
        self.check_input()

        """Get nonzero indices"""
        self.I, self.J = np.where((self.C + self.C.T) > 0.0)

        """Init Vsampler"""
        if self.C.dtype == np.float32:
            self._sampler = RevSampler32(seed)
        elif self.C.dtype == np.float64:
            self._sampler = RevSampler64(seed)
        elif self.C.dtype == np.longdouble:
            self._sampler = RevSampler128(seed)
        else:
            raise ValueError(f"Unknown dtype {self.C.dtype}")

    def check_input(self):
        from deeptime.markov.tools.analysis import is_connected
        if self.C.dtype not in (np.float32, np.float64, np.longdouble):
            raise ValueError("Only supports float32, float64, and longdouble dtype.")
        if not np.all(self.C >= 0):
            raise ValueError("Count matrix contains negative elements")
        if not is_connected(self.C):
            raise ValueError("Count matrix is not connected")
        if not np.all(self.V >= 0.0):
            raise ValueError("P0 contains negative entries")
        if not np.allclose(self.V, self.V.T, atol=1e-6):
            raise ValueError("P0 is not reversible")
        """Check sparsity pattern"""
        iC, jC = np.where((self.C + self.C.T + np.eye(self.C.shape[0])) > 0)
        iV, jV = np.where((self.V + self.V.T + np.eye(self.V.shape[0])) > 0)
        if not np.array_equal(iC, iV):
            raise ValueError('Sparsity patterns of C and X are different.')
        if not np.array_equal(jC, jV):
            raise ValueError('Sparsity patterns of C and X are different.')

    def update(self, N=1):
        self._sampler.update(self.C, self.c, self.V, self.I, self.J, int(N))

    def sample(self, N=1, return_statdist=False):
        self.update(N=N)
        Vsum = self.V.sum(axis=1)
        P = self.V / Vsum[..., np.newaxis]
        if return_statdist:
            nu = self.V.sum(axis=1)
            pi = nu / nu.sum()
            return P, pi
        else:
            return P



from typing import Optional, Union, Callable

import numpy as np

from ._vamp import VAMP
from ..basis import Identity
from ..covariance import Covariance, CovarianceModel
from ..numeric import eig_corr

__author__ = 'marscher, clonker'


class TICA(VAMP):
    r""" Time-lagged independent component analysis (TICA).

    TICA is a linear transformation method. In contrast to PCA, which finds
    coordinates of maximal variance, TICA finds coordinates of maximal
    autocorrelation at the given lag time. Therefore, TICA is useful in order
    to find the *slow* components in a dataset and thus an excellent choice to
    transform molecular dynamics data before clustering data for the
    construction of a Markov model. When the input data is the result of a
    Markov process (such as thermostatted molecular dynamics), TICA finds in
    fact an approximation to the eigenfunctions and eigenvalues of the
    underlying Markov operator :footcite:`perez2013identification`.

    It estimates a TICA transformation from *data*. The resulting model can be used
    to obtain eigenvalues, eigenvectors,
    or project input data onto the slowest TICA components.

    Parameters
    ----------
    lagtime : int or None, optional, default=None
        The lagtime under which covariances are estimated. This is only relevant when estimating from data, in case
        covariances are provided this should either be None or exactly the value that was used to estimate
        said covariances.
    epsilon : float, optional, default=1e-6
        Eigenvalue norm cutoff. Eigenvalues of C0 with norms <= epsilon will be
        cut off. The remaining number of eigenvalues define the size
        of the output.
    dim : int, optional, default=None
        Number of dimensions to keep:

        * if dim is not set (None) all available ranks are kept:
          :code:`n_components == min(n_samples, n_uncorrelated_features)`
        * if dim is an integer >= 1, this number specifies the number
          of dimensions to keep.
    var_cutoff : float, optional, default=None
        Determines the number of output dimensions by including dimensions until their cumulative kinetic variance
        exceeds the fraction subspace variance. var_cutoff=1.0 means all numerically available dimensions
        (see epsilon) will be used, unless set by dim. Setting var_cutoff smaller than 1.0 is exclusive with dim.
    scaling : str or None, default='kinetic_map'
        Can be set to :code:`None`, 'kinetic_map' (:footcite:`noe2015kinetic`),
        or 'commute_map' (:footcite:`noe2016commute`). For more details see :attr:`scaling`.
    observable_transform : callable, optional, default=Identity
        A feature transformation on the raw data which is used to estimate the model.

    Notes
    -----
    Given a sequence of multivariate data :math:`X_t`, it computes the
    mean-free covariance and time-lagged covariance matrix:

    .. math::

        \begin{aligned}
        C_0 &=      (X_t - \mu)^T \mathrm{diag}(w) (X_t - \mu) \\
        C_{\tau} &= (X_t - \mu)^T \mathrm{diag}(w) (X_{t + \tau} - \mu)
        \end{aligned}

    where :math:`w` is a vector of weights for each time step. By default, these weights
    are all equal to one, but different weights are possible, like the re-weighting
    to equilibrium described in :footcite:`wu2017variational`. Subsequently, the eigenvalue problem

    .. math:: C_{\tau} r_i = C_0 \lambda_i r_i,

    is solved,where :math:`r_i` are the independent components and :math:`\lambda_i` are
    their respective normalized time-autocorrelations. The eigenvalues are
    related to the relaxation timescale by

    .. math::

        t_i = -\frac{\tau}{\ln |\lambda_i|}.

    When used as a dimension reduction method, the input data is projected
    onto the dominant independent components.

    Under the assumption of reversible dynamics and the limit of good statistics, the time-lagged autocovariance
    :math:`C_\tau` is symmetric. Due to finite data, this symmetry is explicitly enforced in the estimator.

    TICA was originally introduced for signal processing in :footcite:`molgedey1994separation`. It was
    introduced to molecular dynamics and as a method for the construction
    of Markov models in :footcite:`perez2013identification` and :footcite:`schwantes2013improvements`. It was shown
    in :footcite:`perez2013identification` that when applied
    to molecular dynamics data, TICA is an approximation to the eigenvalues
    and eigenvectors of the true underlying dynamics.

    Examples
    --------
    Invoke TICA transformation with a given lag time and output dimension:

    >>> import numpy as np
    >>> from deeptime.decomposition import TICA
    >>> data = np.random.random((100,3))
    >>> # fixed output dimension
    >>> estimator = TICA(dim=1, lagtime=2).fit(data)
    >>> model_onedim = estimator.fetch_model()
    >>> projected_data = model_onedim.transform(data)
    >>> np.testing.assert_equal(projected_data.shape[1], 1)

    or invoke it with a percentage value of to-be captured kinetic variance (80% in the example)

    >>> estimator = TICA(var_cutoff=0.8, lagtime=2).fit(data)
    >>> model_var = estimator.fetch_model()
    >>> projected_data = model_var.transform(data)

    For a brief explaination why TICA outperforms PCA to extract a good reaction coordinate have a look `here
    <http://docs.markovmodel.org/lecture_tica.html#Example:-TICA-versus-PCA-in-a-stretched-double-well-potential>`_.

    See also
    --------
    CovarianceKoopmanModel : TICA estimation output model

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, lagtime: Optional[int] = None, epsilon: float = 1e-6, dim: Optional[int] = None,
                 var_cutoff: Optional[float] = None, scaling: Optional[str] = 'kinetic_map',
                 observable_transform: Callable[[np.ndarray], np.ndarray] = Identity()):
        super(TICA, self).__init__(lagtime=lagtime, dim=dim, var_cutoff=var_cutoff, scaling=scaling, epsilon=epsilon,
                                   observable_transform=observable_transform)

    @classmethod
    def covariance_estimator(cls, lagtime: int, ncov: Union[int] = float('inf')):
        return Covariance(lagtime=lagtime, compute_c00=True, compute_c0t=True, compute_ctt=False,
                          remove_data_mean=True, reversible=True, bessels_correction=False,
                          ncov=ncov)

    @staticmethod
    def _decomposition(covariances, epsilon, scaling, dim, var_cutoff) -> VAMP._DiagonalizationResults:
        from deeptime.numeric import ZeroRankError

        # diagonalize with low rank approximation
        try:
            eigenvalues, eigenvectors, rank = eig_corr(covariances.cov_00, covariances.cov_0t, epsilon,
                                                       canonical_signs=True, return_rank=True)
        except ZeroRankError:
            raise ZeroRankError('All input features are constant in all time steps. '
                                'No dimension would be left after dimension reduction.')
        if scaling in ('km', 'kinetic_map'):  # scale by eigenvalues
            eigenvectors *= eigenvalues[None, :]
        elif scaling == 'commute_map':  # scale by (regularized) timescales
            lagtime = covariances.lagtime
            timescales = 1. - lagtime / np.log(np.abs(eigenvalues))
            # dampen timescales smaller than the lag time, as in section 2.5 of ref. [5]
            regularized_timescales = 0.5 * timescales * np.maximum(
                np.tanh(np.pi * ((timescales - lagtime) / lagtime) + 1), 0)

            eigenvectors *= np.sqrt(regularized_timescales / 2)

        return VAMP._DiagonalizationResults(
            rank0=rank, rankt=rank, singular_values=eigenvalues,
            left_singular_vecs=eigenvectors, right_singular_vecs=eigenvectors
        )

    @property
    def epsilon(self) -> float:
        r""" Eigenvalue norm cutoff.

        :type: float
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value: float):
        self._epsilon = value

    @property
    def scaling(self) -> Optional[str]:
        r""" Scaling parameter. Can take the following values:

        * None: unscaled.
        * 'kinetic_map': Eigenvectors will be scaled by eigenvalues. As a result, Euclidean
          distances in the transformed data approximate kinetic distances :footcite:`noe2015kinetic`.
          This is a good choice when the data is further processed by clustering.
        * 'commute_map': Eigenvector i will be scaled by sqrt(timescale_i / 2). As a result,
          Euclidean distances in the transformed data will approximate commute distances :footcite:`noe2016commute`.

        :getter: Yields the currently configured scaling.
        :setter: Sets a new scaling.
        :type: str or None
        """
        return self._scaling

    @scaling.setter
    def scaling(self, value: Optional[str]):
        valid_scalings = [None, 'kinetic_map', 'km', 'commute_map']
        if value not in valid_scalings:
            raise ValueError("Scaling parameter is allowed to be one of {}".format(valid_scalings))
        self._scaling = value

    def fit_from_covariances(self, covariances: Union[Covariance, CovarianceModel]):
        r"""Fits a model based on provided symmetrized covariances.

        Parameters
        ----------
        covariances : Covariance or CovarianceModel
            The covariances

        Returns
        -------
        self : TICA
            Reference to self.
        """
        covariances = self._to_covariance_model(covariances)
        if not covariances.symmetrized:
            raise ValueError("The covariance model must be estimated such that the "
                             "autocorrelations are symmetric!")
        if covariances.cov_0t is None:
            raise ValueError("The covariance model must contain C0t!")
        self._model = self._decompose(covariances)
        self._covariance_estimator = None
        return self



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('basis', parent_package, top_path)

    config.add_extension('_basis_bindings',
                         sources=['src/basis_bindings.cpp'],
                         include_dirs=[],
                         language='c++',
                         )
    return config
